from __future__ import annotations
import ast
import inspect
from abc import ABC, abstractmethod
from typing import List, Callable


def py_bitshift(
    left: str | List[str], right: str | List[str], is_downstream: bool = True
):
    """
    >>> render_ast(py_bitshift("foo", "bar", is_downstream=False))
    'foo << bar'
    >>> render_ast(py_bitshift("foo", "bar"))
    'foo >> bar'
    >>> render_ast(py_bitshift("foo", ["bar", "baz"]))
    'foo >> [bar, baz]'
    >>> render_ast(py_bitshift(["foo", "bar"], "baz"))
    '[foo, bar] >> baz'
    """
    left = (
        ast.Name(id=left)
        if isinstance(left, str)
        else ast.List(elts=[ast.Name(id=elt) for elt in left])
    )
    right = (
        ast.Name(id=right)
        if isinstance(right, str)
        else ast.List(elts=[ast.Name(id=elt) for elt in right])
    )
    return ast.Expr(
        value=ast.BinOp(
            left=left, op=ast.RShift() if is_downstream else ast.LShift(), right=right
        )
    )


def py_assigned_object(name: str, obj: str, **kwargs) -> ast.Assign:
    """
    >>> render_ast(py_assigned_object("foo", "Bar", baz="bop"))
    "foo = Bar(baz='bop')"
    """
    return ast.Assign(
        lineno=0,
        targets=[ast.Name(id=name)],
        value=ast.Call(
            func=ast.Name(id=obj),
            args=[],
            keywords=[
                ast.keyword(
                    arg=arg,
                    value=(
                        ast.Constant(value=value)
                        if not isinstance(value, ast.AST)
                        else value
                    ),
                )
                for arg, value in kwargs.items()
            ],
        ),
    )


def py_object(name: str, **kwargs) -> ast.Expr:
    """
    >>> render_ast(py_object("Bar", baz="bop"))
    "Bar(baz='bop')"
    """
    return ast.Expr(
        value=ast.Call(
            func=ast.Name(id=name),
            args=[],
            keywords=[
                ast.keyword(arg=arg, value=ast.Constant(value=value))
                for arg, value in kwargs.items()
            ],
        )
    )


def py_root(*args) -> ast.Module:
    """
    :param args: ast objects, such as ast.Expr
    :return: root ast.Module, which can be `ast.unparse`d
    """
    return ast.Module(body=args, type_ignores=[])


def py_import(
    names: List[str], module: str = None
) -> ast.ImportFrom | ast.Import | list:
    """
    :param module: e.g. `airflow.operators.bash` for `from airflow.operators.bash import BashOperator`
    :param names: e.g. `BashOperator` for `from airflow.operators.bash import BashOperator`
    :return: ast.ImportFrom
    >>> render_ast(py_import(module="airflow.operators.bash", names=["BashOperator"]))
    'from airflow.operators.bash import BashOperator'

    >>> render_ast(py_import(names=["json"]))
    'import json'
    """
    if module is not None:
        return ast.ImportFrom(
            module=module, names=[ast.alias(name=name) for name in names], level=0
        )
    elif module is None and names:
        return ast.Import(names=[ast.alias(name=name) for name in names], level=0)
    else:
        return []


def py_with(
    item: ast.expr, body: List[ast.stmt], assignment: str | None = None
) -> ast.With:
    # noinspection PyTypeChecker
    """
    >>> render_ast(py_with(py_object("Bar"), [ast.Pass()]))
    'with Bar():\\n    pass'
    >>> render_ast(
    ...   py_with(py_object("DAG", dag_id="foo").value, [py_object("Operator", task_id="foo")])
    ... )
    "with DAG(dag_id='foo'):\\n    Operator(task_id='foo')"
    >>> render_ast(
    ...   py_with(py_object("DAG", dag_id="foo").value, [py_object("Operator", task_id="foo")], "dag")
    ... )
    "with DAG(dag_id='foo') as dag:\\n    Operator(task_id='foo')"
    """
    if isinstance(item, ast.Expr):
        item = item.value
    return ast.With(
        items=[
            ast.withitem(
                context_expr=item,
                optional_vars=[ast.Name(id=assignment)] if assignment else [],
            )
        ],
        body=body,
        lineno=1,
    )


def py_function(c: Callable):
    """
    >>> def foo(a, b):
    ...    print(a + b)
    >>> render_ast(py_function(foo))
    'def foo(a, b):\\n    print(a + b)'
    """
    return ast.parse(inspect.getsource(c)).body[0]


def py_reference(name: str) -> ast.Expr:
    """
    ```pycon
    >>> render_ast(py_reference("foo"))
    'foo'

    ```
    """
    return ast.Expr(value=ast.Name(id=name))


def render_ast(ast_object) -> str:
    return ast.unparse(ast_object)


class OrbiterASTBase(ABC):
    @abstractmethod
    def _to_ast(self) -> ast.stmt | ast.Module:
        raise NotImplementedError()

    def __str__(self):
        return render_ast(self._to_ast())

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __gt__(self, other):
        return str(self) > str(other)

    def __hash__(self):
        return hash(str(self))



from orbiter.objects import ImportList
from orbiter.objects.requirement import OrbiterRequirement
from orbiter.objects.task import OrbiterOperator, RenderAttributes

__mermaid__ = """
--8<-- [start:mermaid-relationships]
OrbiterOperator  "implements" <|--  OrbiterEmptyOperator
--8<-- [end:mermaid-relationships]
"""


class OrbiterEmptyOperator(OrbiterOperator):
    """
    An Airflow EmptyOperator. Does nothing.

    ```pycon
    >>> OrbiterEmptyOperator(task_id="foo")
    foo_task = EmptyOperator(task_id='foo')

    ```
    :param task_id: The `task_id` for the operator
    :type task_id: str
    :param **kwargs: Extra arguments to pass to the operator
    :param **OrbiterBase: [OrbiterBase][orbiter.objects.OrbiterBase] inherited properties
    """

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    operator = "BashOperator"
    task_id: str
    --8<-- [end:mermaid-props]
    """

    imports: ImportList = [
        OrbiterRequirement(
            package="apache-airflow",
            module="airflow.operators.empty",
            names=["EmptyOperator"],
        )
    ]
    operator: str = "EmptyOperator"
    # noinspection Pydantic
    render_attributes: RenderAttributes = OrbiterOperator.render_attributes



from __future__ import annotations

from abc import ABC
from typing import List, Annotated, Set

from pydantic import BaseModel, AfterValidator
from orbiter.objects.requirement import OrbiterRequirement
from orbiter.objects.connection import OrbiterConnection
from orbiter.objects.env_var import OrbiterEnvVar
from orbiter.objects.variable import OrbiterVariable
from orbiter.objects.include import OrbiterInclude


CALLBACK_KEYS = [
    "on_success_callback",
    "on_failure_callback",
    "sla_miss_callback",
    "on_retry_callback",
    "on_execute_callback",
    "on_skipped_callback",
]


def validate_imports(v):
    assert len(v)
    for i in v:
        assert isinstance(i, OrbiterRequirement)
    return v


ImportList = Annotated[List[OrbiterRequirement], AfterValidator(validate_imports)]


class OrbiterBase(BaseModel, ABC, arbitrary_types_allowed=True):
    """**AbstractBaseClass** for Orbiter objects, provides a number of properties

    :param imports: List of [OrbiterRequirement][orbiter.objects.requirement.OrbiterRequirement] objects
    :type imports: List[OrbiterRequirement]
    :param orbiter_kwargs: Optional dictionary of keyword arguments, to preserve what was originally parsed by a rule
    :type orbiter_kwargs: dict, optional
    :param orbiter_conns: Optional set of [OrbiterConnection][orbiter.objects.connection.OrbiterConnection] objects
    :type orbiter_conns: Set[OrbiterConnection], optional
    :param orbiter_env_vars: Optional set of [OrbiterEnvVar][orbiter.objects.env_var.OrbiterEnvVar] objects
    :type orbiter_env_vars: Set[OrbiterEnvVar], optional
    :param orbiter_includes: Optional set of [OrbiterInclude][orbiter.objects.include.OrbiterInclude] objects
    :type orbiter_includes: Set[OrbiterInclude], optional
    :param orbiter_vars: Optional set of [OrbiterVariable][orbiter.objects.variable.OrbiterVariable] objects
    :type orbiter_vars: Set[OrbiterVariable], optional
    """

    imports: ImportList
    orbiter_kwargs: dict = None

    orbiter_conns: Set[OrbiterConnection] | None = None
    orbiter_env_vars: Set[OrbiterEnvVar] | None = None
    orbiter_includes: Set[OrbiterInclude] | None = None
    orbiter_vars: Set[OrbiterVariable] | None = None


def conn_id(conn_id: str, prefix: str = "", conn_type: str = "generic") -> dict:
    """Helper function to add an [OrbiterConnection][orbiter.objects.connection.OrbiterConnection]
    when adding a `conn_id`

    Usage:
    ```python
    OrbiterBashOperator(
        **conn_id("my_conn_id")
    )
    ```
    :param conn_id: The connection id
    :type conn_id: str
    :param prefix: Prefix to add to the connection id
    :type prefix: str, optional
    :param conn_type: Connection type
    :type conn_type: str, optional
    :return: Dictionary to unpack (e.g. `**conn_id(...)`)
    :rtype: dict
    """
    from orbiter.objects.connection import OrbiterConnection

    return {
        f"{prefix + '_' if prefix else ''}conn_id": conn_id,
        "orbiter_conns": {
            OrbiterConnection(
                conn_id=conn_id, **({"conn_type": conn_type} if conn_type else {})
            )
        },
    }


def pool(name: str) -> dict:
    """Helper function to add a [OrbiterPool][orbiter.objects.pool.OrbiterPool] when adding a `pool`

    Usage:
    ```python
    OrbiterBashOperator(
        **pool("my_pool")
    )
    ```
    :param name: The pool name
    :type name: str
    :return: Dictionary to unpack (e.g. `**pool(...)`)
    :rtype: dict
    """
    from orbiter.objects.pool import OrbiterPool

    return {
        "pool": name,
        "orbiter_pools": {OrbiterPool(name=name)},
    }



from typing import ClassVar

from pydantic import BaseModel

from orbiter.objects.airflow_settings import AirflowSettingsRender


class OrbiterPool(BaseModel, AirflowSettingsRender, extra="forbid"):
    """An Airflow
    [Pool](https://airflow.apache.org/docs/apache-airflow/stable/administration-and-deployment/pools.html),
    rendered to an
    [`airflow_settings.yaml`](https://www.astronomer.io/docs/astro/cli/develop-project#configure-airflow_settingsyaml-local-development-only)
    file.


    ```pycon
    >>> OrbiterPool(name="foo", description="bar", slots=5).render()
    {'pool_name': 'foo', 'pool_description': 'bar', 'pool_slot': 5}

    ```

    !!! note

        Use the utility `pool` function to easily generate both an `OrbiterPool`
        and `pool` property for an operator

        ```python
        from orbiter.objects import pool

        OrbiterTask(
            ... ,
            **pool("my_pool"),
        )
        ```

    :param name: The name of the pool
    :type name: str
    :param description: The description of the pool
    :type description: str, optional
    :param slots: The number of slots in the pool. Defaults to 128
    :type slots: int, optional
    """  # noqa: E501

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    name: str
    description: str | None
    slots: int | None
    --8<-- [end:mermaid-props]
    """

    render_prefix: ClassVar[str] = "pool_"

    name: str
    description: str = ""
    slots: int = 128

    def __add__(self, other) -> "OrbiterPool":
        self.slots = max((self.slots, other.slots))
        self.description = max((self.description, other.description))
        return self

    def __hash__(self):
        return hash(f"{self.model_dump_json}")

    def render(self) -> dict:
        # airflow_settings.yaml takes 'slot' not 'slots'
        res = super().render()
        res[self.render_prefix + "slot"] = res[self.render_prefix + "slots"]
        del res[self.render_prefix + "slots"]
        return res



from __future__ import annotations

import json
from abc import ABC
from functools import partial
from typing import Callable, Set, ClassVar, Any

import xmltodict
import yaml
from pydantic import (
    BaseModel,
)
from pydantic.v1 import validator


class FileType(BaseModel, ABC, arbitrary_types_allowed=True):
    """**Abstract Base** File Type

    :param extension: The file extension(s) for this file type
    :type extension: Set[str]
    :param load_fn: The function to load the file into a dictionary for this file type
    :type load_fn: Callable[[str], dict]
    :param dump_fn: The function to dump a dictionary to a string for this file type
    :type dump_fn: Callable[[dict], str]
    """

    extension: ClassVar[Set[str]]
    load_fn: ClassVar[Callable[[str], dict]]
    dump_fn: ClassVar[Callable[[dict], str]]

    def __hash__(self):
        return hash(tuple(self.extension))

    @validator("extension", pre=True)
    @classmethod
    def ext_validate(cls, v: Set[str]):
        if not v:
            raise ValueError("Extension cannot be an empty set")
        for ext in v:
            if not isinstance(ext, str):
                raise ValueError("Extension should be a string")
            if "." in v:
                raise ValueError("Extension should not contain '.'")
        return {ext.lower() for ext in v}


class FileTypeJSON(FileType):
    """JSON File Type

    ```pycon
    >>> out = FileTypeJSON.dump_fn({'a': 1}); out
    '{"a": 1}'
    >>> FileTypeJSON.load_fn(out)
    {'a': 1}

    ```
    :param extension: JSON
    :type extension: Set[str]
    :param load_fn: json.loads
    :type load_fn: Callable[[str], dict]
    :param dump_fn: json.dumps
    :type dump_fn: Callable[[dict], str]
    """

    extension: ClassVar[Set[str]] = {"JSON"}
    load_fn: ClassVar[Callable[[str], dict]] = json.loads
    dump_fn: ClassVar[Callable[[dict], str]] = partial(json.dumps, default=str)


# noinspection t
def xmltodict_parse(input_str: str) -> Any:
    """Calls `xmltodict.parse` and does post-processing fixes.

    !!! note

        The original [`xmltodict.parse`](https://pypi.org/project/xmltodict/) method returns EITHER:

        - a dict (one child element of type)
        - or a list of dict (many child element of type)

        This behavior can be confusing, and is an issue with the original xml spec being referenced.

        **This method deviates by standardizing to the latter case (always a `list[dict]`).**

        **All XML elements will be a list of dictionaries, even if there's only one element.**

    ```pycon
    >>> xmltodict_parse("")
    Traceback (most recent call last):
    xml.parsers.expat.ExpatError: no element found: line 1, column 0
    >>> xmltodict_parse("<a></a>")
    {'a': None}
    >>> xmltodict_parse("<a foo='bar'></a>")
    {'a': [{'@foo': 'bar'}]}
    >>> xmltodict_parse("<a foo='bar'><foo bar='baz'></foo></a>")  # Singleton - gets modified
    {'a': [{'@foo': 'bar', 'foo': [{'@bar': 'baz'}]}]}
    >>> xmltodict_parse("<a foo='bar'><foo bar='baz'><bar><bop></bop></bar></foo></a>")  # Nested Singletons - modified
    {'a': [{'@foo': 'bar', 'foo': [{'@bar': 'baz', 'bar': [{'bop': None}]}]}]}
    >>> xmltodict_parse("<a foo='bar'><foo bar='baz'></foo><foo bing='bop'></foo></a>")
    {'a': [{'@foo': 'bar', 'foo': [{'@bar': 'baz'}, {'@bing': 'bop'}]}]}

    ```
    :param input_str: The XML string to parse
    :type input_str: str
    :return: The parsed XML
    :rtype: dict
    """

    # noinspection t
    def _fix(d):
        """fix the dict in place, recursively, standardizing on a list of dict even if there's only one entry."""
        # if it's a dict, descend to fix
        if isinstance(d, dict):
            for k, v in d.items():
                # @keys are properties of elements, non-@keys are elements
                if not k.startswith("@"):
                    if isinstance(v, dict):
                        # THE FIX
                        # any non-@keys should be a list of dict, even if there's just one of the element
                        d[k] = [v]
                        _fix(v)
                    else:
                        _fix(v)
        # if it's a list, descend to fix
        if isinstance(d, list):
            for v in d:
                _fix(v)

    output = xmltodict.parse(input_str)
    _fix(output)
    return output


class FileTypeXML(FileType):
    """XML File Type

    !!! note
        This class uses a custom `xmltodict_parse` method to standardize the output to a list of dictionaries

    ```pycon
    >>> out = FileTypeXML.dump_fn({'a': 1}); out
    '<?xml version="1.0" encoding="utf-8"?>\\n<a>1</a>'
    >>> FileTypeXML.load_fn(out)
    {'a': '1'}

    ```
    :param extension: XML
    :type extension: Set[str]
    :param load_fn: xmltodict_parse
    :type load_fn: Callable[[str], dict]
    :param dump_fn: xmltodict.unparse
    :type dump_fn: Callable[[dict], str]
    """

    extension: ClassVar[Set[str]] = {"XML"}
    load_fn: ClassVar[Callable[[str], dict]] = xmltodict_parse
    dump_fn: ClassVar[Callable[[dict], str]] = xmltodict.unparse


class FileTypeYAML(FileType):
    """YAML File Type


    ```pycon
    >>> out = FileTypeYAML.dump_fn({'a': 1}); out
    'a: 1\\n'
    >>> FileTypeYAML.load_fn(out)
    {'a': 1}

    ```
    :param extension: YAML, YML
    :type extension: Set[str]
    :param load_fn: yaml.safe_load
    :type load_fn: Callable[[str], dict]
    :param dump_fn: yaml.safe_dump
    :type dump_fn: Callable[[dict], str]
    """

    extension: ClassVar[Set[str]] = {"YAML", "YML"}
    load_fn: ClassVar[Callable[[str], dict]] = yaml.safe_load
    dump_fn: ClassVar[Callable[[dict], str]] = yaml.safe_dump


if __name__ == "__main__":
    import doctest

    doctest.testmod(
        optionflags=doctest.ELLIPSIS
        | doctest.NORMALIZE_WHITESPACE
        | doctest.IGNORE_EXCEPTION_DETAIL
    )



from pydantic import BaseModel


class OrbiterInclude(BaseModel, extra="forbid"):
    """Represents an included file in an `/include` directory

    :param filepath: The relative path (from the output directory) to write the file to
    :type filepath: str
    :param contents: The contents of the file
    :type contents: str
    """

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    filepath: str
    contents: str
    --8<-- [end:mermaid-props]
    """

    filepath: str
    contents: str

    def __hash__(self):
        return hash(self.contents + self.filepath)

    def render(self):
        return self.contents



from __future__ import annotations

from typing import Set

from orbiter.objects import OrbiterRequirement, ImportList, OrbiterConnection
from orbiter.objects.task import OrbiterOperator, RenderAttributes

__mermaid__ = """
--8<-- [start:mermaid-relationships]
OrbiterOperator "implements" <|-- OrbiterEmailOperator
--8<-- [end:mermaid-relationships]
"""


class OrbiterEmailOperator(OrbiterOperator):
    # noinspection GrazieInspection
    """
    An Airflow
    [EmailOperator](https://registry.astronomer.io/providers/apache-airflow/versions/latest/modules/EmailOperator).
    Used to send emails.

    ```pycon
    >>> OrbiterEmailOperator(
    ...   task_id="foo", to="humans@astronomer.io", subject="Hello", html_content="World!"
    ... )
    foo_task = EmailOperator(task_id='foo', to='humans@astronomer.io', subject='Hello', html_content='World!', conn_id='SMTP')

    ```
    :param task_id: The `task_id` for the operator
    :type task_id: str
    :param to: The recipient of the email
    :type to: str | list[str]
    :param subject: The subject of the email
    :type subject: str
    :param html_content: The content of the email
    :type html_content: str
    :param files: The files to attach to the email, defaults to None
    :type files: list, optional
    :param conn_id: The SMTP connection to use. Defaults to "SMTP" and sets `orbiter_conns` property
    :type conn_id: str, optional
    :param **kwargs: Extra arguments to pass to the operator
    :param **OrbiterBase: [OrbiterBase][orbiter.objects.OrbiterBase] inherited properties
    """  # noqa: E501

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    operator = "EmailOperator"
    task_id: str
    to: str | list[str]
    subject: str
    html_content: str
    files: list | None
    conn_id: str
    --8<-- [end:mermaid-props]
    """

    imports: ImportList = [
        OrbiterRequirement(
            package="apache-airflow-providers-smtp",
            module="airflow.providers.smtp.operators.smtp",
            names=["EmailOperator"],
        )
    ]
    operator: str = "EmailOperator"
    # noinspection Pydantic
    render_attributes: RenderAttributes = OrbiterOperator.render_attributes + [
        "to",
        "subject",
        "html_content",
        "files",
        "conn_id",
    ]
    to: str | list[str]
    subject: str
    html_content: str
    files: list | None = []
    conn_id: str | None = "SMTP"
    orbiter_conns: Set[OrbiterConnection] | None = {
        OrbiterConnection(conn_id="SMTP", conn_type="smtp")
    }



"""
The brain of the Orbiter framework is in it's [`Rules`][orbiter.rules.Rule]
and the [`Rulesets`][orbiter.rules.rulesets.Ruleset] that contain them.

- A [`Rule`][orbiter.rules.Rule] is a python function that is evaluated and produces **something**
(typically an [Object](../objects/index.md)) or **nothing**
- A [`Ruleset`][orbiter.rules.rulesets.Ruleset] is a collection of [`Rules`][orbiter.rules.Rule] that are
    evaluated in priority order
- A [`TranslationRuleset`][orbiter.rules.rulesets.TranslationRuleset]
    is a collection of [`Rulesets`][orbiter.rules.rulesets.Ruleset],
    relating to an [Origin](../origins.md) and [`FileType`][orbiter.file_types.FileType],
    with a [`translate_fn`][orbiter.rules.rulesets.translate] which determines how to apply the rulesets.

Different [`Rules`][orbiter.rules.Rule] are applied in different scenarios, such as:

- converting input to an Airflow DAG ([`@dag_rule`][orbiter.rules.DAGRule]),
- converting input to a specific Airflow Operator ([`@task_rule`][orbiter.rules.TaskRule]),
- filtering entries from the input data
([`@dag_filter_rule`][orbiter.rules.DAGFilterRule], [`@task_filter_rule`][orbiter.rules.TaskFilterRule]).

!!! tip

    To map the following input
    ```json
    {
        "id": "my_task",
        "command": "echo 'hi'"
    }
    ```

    to an Airflow
    [BashOperator](https://registry.astronomer.io/providers/apache-airflow/versions/latest/modules/BashOperator),
    a [`Rule`][orbiter.rules.Rule] could parse it as follows:

    ```python
    @task_rule
    def my_rule(val):
        if 'command' in val:
            return OrbiterBashOperator(task_id=val['id'], bash_command=val['command'])
        else:
            return None
    ```

    This returns a
    [`OrbiterBashOperator`][orbiter.objects.operators.bash.OrbiterBashOperator], which will become an Airflow
    [BashOperator](https://registry.astronomer.io/providers/apache-airflow/versions/latest/modules/BashOperator)
    when the translation completes.
"""

from __future__ import annotations

import functools
import json
import re
from typing import Callable, Any, Collection, TYPE_CHECKING, List, Mapping

from pydantic import BaseModel, Field

from loguru import logger

from orbiter.config import TRIM_LOG_OBJECT_LENGTH
from orbiter.objects.task import OrbiterOperator, OrbiterTaskDependency

if TYPE_CHECKING:
    from orbiter.objects.task_group import OrbiterTaskGroup
    from orbiter.objects.project import OrbiterProject
    from orbiter.objects.dag import OrbiterDAG

qualname_validator_regex = r"^[\w.]+$"
qualname_validator = re.compile(qualname_validator_regex)


def trim_dict(v):
    """Stringify and trim a dictionary if it's greater than a certain length
    (used to trim down overwhelming log output)"""
    if TRIM_LOG_OBJECT_LENGTH != -1 and isinstance(v, Mapping):
        if len(str(v)) > TRIM_LOG_OBJECT_LENGTH:
            return json.dumps(v, default=str)[:TRIM_LOG_OBJECT_LENGTH] + "..."
    if isinstance(v, list):
        return [trim_dict(_v) for _v in v]
    return v


def rule(
    func=None, *, priority=None
) -> (
    Rule
    | "DAGFilterRule"
    | "DAGRule"
    | "TaskFilterRule"
    | "TaskRule"
    | "TaskDependencyRule"
    | "PostProcessingRule"
):
    if func is None:
        return functools.partial(rule, priority=priority)

    priority = priority or 1
    _rule = Rule(priority=priority, rule=func)
    functools.update_wrapper(_rule, func)
    return _rule


class Rule(BaseModel, Callable, extra="forbid"):
    """
    A `Rule` contains a python function that is evaluated and produces something
    (typically an [Object](../objects/index.md)) or nothing

    A `Rule` can be created from a decorator
    ```pycon
    >>> @rule(priority=1)
    ... def my_rule(val):
    ...     return 1
    >>> isinstance(my_rule, Rule)
    True
    >>> my_rule(val={})
    1

    ```

    The function in a rule takes one parameter (`val`), and **must always evaluate to *something* or *nothing*.**
    ```pycon
    >>> Rule(rule=lambda val: 4)({})
    4
    >>> Rule(rule=lambda val: None)({})

    ```

    !!! tip

        If the returned value is an [Orbiter Object](../../objects/index.md),
        the passed `kwargs` are saved in a special `orbiter_kwargs` property

        ```pycon
        >>> from orbiter.objects.dag import OrbiterDAG
        >>> @rule
        ... def my_rule(foo):
        ...     return OrbiterDAG(dag_id="", file_path="")
        >>> my_rule(foo="bar").orbiter_kwargs
        {'foo': 'bar'}

        ```

    !!! note

        A `Rule` must have a `rule` property and extra properties cannot be passed

        ```pycon
        >>> # noinspection Pydantic
        ... Rule(rule=lambda: None, not_a_prop="???")
        ... # doctest: +ELLIPSIS
        Traceback (most recent call last):
        pydantic_core._pydantic_core.ValidationError: ...

        ```

    :param rule: Python function to evaluate. Takes a single argument and returns **something** or **nothing**
    :type rule: Callable[[dict | Any], Any | None]
    :param priority: Higher priority rules are evaluated first, must be greater than 0. Default is 0
    :type priority: int, optional
    """

    rule: Callable[[dict | Any], Any | None]
    priority: int = Field(0, ge=0)

    def __call__(self, *args, **kwargs):
        try:
            result = self.rule(*args, **kwargs)
            # Save the original kwargs under orbiter_kwargs
            if result:
                if kwargs and hasattr(result, "orbiter_kwargs"):
                    setattr(result, "orbiter_kwargs", kwargs)
        except Exception as e:
            logger.warning(
                f"[RULE]: {self.rule.__name__}\n"
                f"[ERROR]:\n{type(e)} - {trim_dict(e)}\n"
                f"[INPUT]:\n{trim_dict(args)}\n{trim_dict(kwargs)}"
            )
            result = None
        return result


class DAGFilterRule(Rule):
    """The `@dag_filter_rule` decorator creates a [`DAGFilterRule`][orbiter.rules.DAGFilterRule]

    ```python
    @dag_filter_rule
    def foo(val: dict) -> List[dict]:
        return [{"dag_id": "foo"}]
    ```

    !!! hint

        In addition to filtering, a `DAGFilterRule` can also map input to a more reasonable output for later processing
    """

    rule: Callable[[dict], Collection[dict] | None]


dag_filter_rule: Callable[[...], DAGFilterRule] = rule


class DAGRule(Rule):
    """A `@dag_rule` decorator creates a [`DAGRule`][orbiter.rules.DAGRule]

    !!! tip

        A `__file` key is added to the original input, which is the file path of the input.

    ```python
    @dag_rule
    def foo(val: dict) -> OrbiterDAG | None:
        if 'id' in val:
            return OrbiterDAG(dag_id=val["id"], file_path=f"{val["id"]}.py")
        else:
            return None
    ```
    """

    rule: Callable[[dict], OrbiterDAG | None]


dag_rule: Callable[[...], DAGRule] = rule


class TaskFilterRule(Rule):
    # noinspection PyUnresolvedReferences
    """A `@task_filter_rule` decorator creates a [`TaskFilterRule`][orbiter.rules.TaskFilterRule]

    ```python
    @task_filter_rule
    def foo(val: dict) -> List[dict] | None:
        return [{"task_id": "foo"}]
    ```

    !!! hint

        In addition to filtering, a `TaskFilterRule` can also map input to a more reasonable output for later processing

    :param val: A dictionary of the task
    :type val: dict
    :return: A list of dictionaries of possible tasks or `None`
    :rtype: List[dict] | None
    """
    rule: Callable[[dict], Collection[dict] | None]


task_filter_rule: Callable[[...], TaskFilterRule] = rule


class TaskRule(Rule):
    # noinspection PyUnresolvedReferences
    """A `@task_rule` decorator creates a [`TaskRule`][orbiter.rules.TaskRule]

    ```python
    @task_rule
    def foo(val: dict) -> OrbiterOperator | OrbiterTaskGroup:
        if 'id' in val and 'command' in val:
            return OrbiterBashOperator(task_id=val['id'], bash_command=val['command'])
        else:
            return None
    ```

    :param val: A dictionary of the task
    :type val: dict
    :return: A subclass of [`OrbiterOperator`][orbiter.objects.task.OrbiterOperator]
        or [`OrbiterTaskGroup`][orbiter.objects.task_group.OrbiterTaskGroup] or `None`
    :rtype: OrbiterOperator | OrbiterTaskGroup | None
    """
    rule: Callable[[dict], OrbiterOperator | OrbiterTaskGroup | None]


task_rule: Callable[[...], TaskRule] = rule


class TaskDependencyRule(Rule):
    # noinspection PyUnresolvedReferences
    """
    An `@task_dependency_rule` decorator creates a [`TaskDependencyRule`][orbiter.rules.TaskDependencyRule],
    which takes an [`OrbiterDAG`][orbiter.objects.dag.OrbiterDAG]
    and returns a [`list[OrbiterTaskDependency]`][orbiter.objects.task.OrbiterTaskDependency] or `None`

    ```python
    @task_dependency_rule
    def foo(val: OrbiterDAG) -> OrbiterTaskDependency:
        return [OrbiterTaskDependency(task_id="upstream", downstream="downstream")]
    ```

    :param val: An [`OrbiterDAG`][orbiter.objects.dag.OrbiterDAG]
    :type val: OrbiterDAG
    :return: A list of [`OrbiterTaskDependency`][orbiter.objects.task.OrbiterTaskDependency] or `None`
    :rtype: List[OrbiterTaskDependency] | None
    """
    rule: Callable[[OrbiterDAG], List[OrbiterTaskDependency] | None]


task_dependency_rule: Callable[[...], TaskDependencyRule] = rule


class PostProcessingRule(Rule):
    # noinspection PyUnresolvedReferences
    """
    An `@post_processing_rule` decorator creates a [`PostProcessingRule`][orbiter.rules.PostProcessingRule],
    which takes an [`OrbiterProject`][orbiter.objects.project.OrbiterProject],
    after all other rules have been applied, and modifies it in-place.

    ```python
    @post_processing_rule
    def foo(val: OrbiterProject) -> None:
        val.dags["foo"].tasks["bar"].doc = "Hello World"
    ```

    :param val: An [`OrbiterProject`][orbiter.objects.project.OrbiterProject]
    :type val: OrbiterProject
    :return: `None`
    :rtype: None
    """
    rule: Callable[[OrbiterProject], None]


post_processing_rule: Callable[[...], PostProcessingRule] = rule


@task_rule(priority=1)
def cannot_map_rule(val: dict) -> OrbiterOperator | None:
    """Can be used in a TaskRuleset.
    Returns an `OrbiterEmptyOperator` with a doc string that says it cannot map the task.
    Useful to ensure that tasks that cannot be mapped are still visible in the output.
    """
    from orbiter.objects.operators.empty import OrbiterEmptyOperator

    # noinspection PyArgumentList
    return OrbiterEmptyOperator(
        task_id="UNKNOWN",
        doc_md=f"""[task_type=UNKNOWN] Input did not translate: `{trim_dict(val)}`""",
    )


EMPTY_RULE = Rule(rule=lambda _: None, priority=0)
"""Empty rule, for testing"""


if __name__ == "__main__":
    import doctest

    doctest.testmod(
        optionflags=doctest.ELLIPSIS
        | doctest.NORMALIZE_WHITESPACE
        | doctest.IGNORE_EXCEPTION_DETAIL
    )



from __future__ import annotations

import re
from typing import Any, Tuple

__version__ = "1.2.2"

version = __version__


def clean_value(s: str):
    """Cleans a string to be a standard value, such as one that might be a python variable name

    Contains similar logic to
    https://github.com/apache/airflow/blob/279d1f7c6483910906a567d8416cd2f230c9df31/airflow/utils/helpers.py#L50

    - Changes - to _
    - Changes spaces to _
    - Strips anything non-alphanumeric and not _ or .
    - Adds an "n" prefix if it starts with a number.
    - snake_cases the string

    ```pycon
    >>> clean_value("MyTaskId")
    'my_task_id'
    >>> clean_value("my-Task ID")
    'my_task_id'
    >>> clean_value("01 My Task")
    'n01_my_task'

    ```
    """
    import inflection

    s = s[:250]  # shorten the string to 250 chars max
    s = inflection.underscore(s)  # swap to snake_case
    s = re.sub("[ -]", "_", s)  # swap space or dash to _
    s = re.sub("[^A-Za-z0-9_.]", "", s)  # remove anything that isn't alphanum or _ or .
    s = re.sub(r"^(\d)", r"n\1", s)  # prefix a number at the start with an "n"
    return s


def import_from_qualname(qualname) -> Tuple[str, Any]:
    """Import a function or module from a qualified name
    :param qualname: The qualified name of the function or module to import (e.g. a.b.d.MyOperator or json)
    :return Tuple[str, Any]: The name of the function or module, and the function or module itself
    >>> import_from_qualname('json.loads')
    ('loads', <function loads at ...>)
    >>> import_from_qualname('json')
    ('json', <module 'json' from '...'>)
    """
    from importlib import import_module

    [module, name] = (
        qualname.rsplit(".", 1) if "." in qualname else [qualname, qualname]
    )
    imported_module = import_module(module)
    return (
        name,
        getattr(imported_module, name) if "." in qualname else imported_module,
    )


if __name__ == "__main__":
    import doctest

    doctest.testmod(
        optionflags=doctest.ELLIPSIS
        | doctest.NORMALIZE_WHITESPACE
        | doctest.IGNORE_EXCEPTION_DETAIL
    )



import ast

from orbiter.ast_helper import OrbiterASTBase, py_object, py_reference
from orbiter.objects import OrbiterBase, ImportList, OrbiterRequirement
from orbiter.objects.task import RenderAttributes


__mermaid__ = """
--8<-- [start:mermaid-relationships]
--8<-- [end:mermaid-relationships]
"""


class OrbiterCallback(OrbiterASTBase, OrbiterBase, extra="forbid"):
    """Represents an Airflow
    [callback function](https://airflow.apache.org/docs/apache-airflow/stable/administration-and-deployment/logging-monitoring/callbacks.html),
    which might be used in `DAG.on_failure_callback`, or `Task.on_success_callback`, or etc.

    Can be instantiated directly as a bare callback function (with no arguments):
    ```pycon
    >>> from orbiter.objects.dag import OrbiterDAG
    >>> from orbiter.objects.include import OrbiterInclude
    >>> my_callback = OrbiterCallback(
    ...     function="my_callback",
    ...     imports=[OrbiterRequirement(module="my_callback", names=["my_callback"])],
    ...     orbiter_includes={OrbiterInclude(filepath="my_callback.py", contents="...")}
    ... )
    >>> OrbiterDAG(dag_id='', file_path='', on_failure_callback=my_callback)
    ... # doctest: +ELLIPSIS
    from airflow import DAG
    from my_callback import my_callback
    ...
    with DAG(... on_failure_callback=my_callback):

    ```

    or be subclassed:
    ```pycon
    >>> class OrbiterMyCallback(OrbiterCallback):
    ...   function: str = "my_callback"
    ...   foo: str
    ...   bar: str
    ...   render_attributes: RenderAttributes = ["foo", "bar"]
    >>> OrbiterMyCallback(foo="fop", bar="bop")
    my_callback(foo='fop', bar='bop')

    ```
    :param function: The name of the function to call
    :type function: str
    :param **OrbiterBase: [OrbiterBase][orbiter.objects.OrbiterBase] inherited properties
    """  # noqa: E501

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    function: str
    --8<-- [end:mermaid-props]
    """

    imports: ImportList = [OrbiterRequirement(package="apache-airflow")]
    function: str
    render_attributes: RenderAttributes = []

    def _to_ast(self) -> ast.Expr:
        if self.render_attributes:
            return py_object(
                name=self.function,
                **{
                    k: getattr(self, k)
                    for k in self.render_attributes
                    if (k and getattr(self, k)) or (k == "from_email")
                },
            )
        else:
            return py_reference(self.function)



from __future__ import annotations

import ast

from pydantic import BaseModel

from orbiter.ast_helper import OrbiterASTBase, py_object
from orbiter.objects import OrbiterBase
from orbiter.objects.task import RenderAttributes


class OrbiterTimetable(OrbiterBase, OrbiterASTBase, BaseModel, extra="allow"):
    """An [Airflow Timetable](https://airflow.apache.org/docs/apache-airflow/stable/authoring-and-scheduling/timetable.html)
    reference.

    Utilizes [`OrbiterInclude`][orbiter.objects.include.OrbiterInclude]
    to add a file to a /plugins folder to register the timetable.

    :param **kwargs: any other kwargs to provide to Timetable
    """  # noqa: E501

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    imports: List[OrbiterRequirements]
    orbiter_includes: Set[OrbiterIncludes]
    **kwargs: dict
    --8<-- [end:mermaid-props]
    """

    render_attributes: RenderAttributes = []

    def _to_ast(self) -> ast.stmt | ast.Module:
        # Figure out which timetable we are talking about
        timetable_names = [
            name
            for _import in self.imports
            for name in _import.names
            if "timetable" in name.lower()
        ]
        if len(timetable_names) != 1:
            raise ValueError(
                f"Expected exactly one Timetable name, got {timetable_names}"
            )
        [timetable] = timetable_names

        return py_object(
            timetable,
            **{
                k: getattr(self, k)
                for k in self.render_attributes
                if k and getattr(self, k)
            },
            **{k: getattr(self, k) for k in (self.model_extra.keys() or [])},
        )



from orbiter.objects import ImportList
from orbiter.objects.requirement import OrbiterRequirement
from orbiter.objects.task import OrbiterOperator, RenderAttributes

__mermaid__ = """
--8<-- [start:mermaid-relationships]
OrbiterOperator "implements" <|-- OrbiterBashOperator
--8<-- [end:mermaid-relationships]
"""


class OrbiterBashOperator(OrbiterOperator):
    """
    An Airflow
    [BashOperator](https://registry.astronomer.io/providers/apache-airflow/versions/latest/modules/BashOperator).
    Used to run shell commands.

    ```pycon
    >>> OrbiterBashOperator(task_id="foo", bash_command="echo 'hello world'")
    foo_task = BashOperator(task_id='foo', bash_command="echo 'hello world'")

    ```
    :param task_id: The `task_id` for the operator
    :type task_id: str
    :param bash_command: The shell command to execute
    :type bash_command: str
    :param **kwargs: Extra arguments to pass to the operator
    :param **OrbiterBase: [OrbiterBase][orbiter.objects.OrbiterBase] inherited properties
    """

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    operator = "BashOperator"
    task_id: str
    bash_command: str
    --8<-- [end:mermaid-props]
    """

    imports: ImportList = [
        OrbiterRequirement(
            package="apache-airflow",
            module="airflow.operators.bash",
            names=["BashOperator"],
        )
    ]
    operator: str = "BashOperator"
    # noinspection Pydantic
    render_attributes: RenderAttributes = OrbiterOperator.render_attributes + [
        "bash_command"
    ]

    bash_command: str



from __future__ import annotations

import sys
import re
from functools import reduce
from pathlib import Path
from typing import Dict, Iterable, Set, Literal

import yaml
from loguru import logger
from pydantic import validate_call

from orbiter.objects.callbacks import OrbiterCallback
from orbiter.objects.connection import OrbiterConnection
from orbiter.objects.dag import OrbiterDAG
from orbiter.objects.env_var import OrbiterEnvVar
from orbiter.objects.include import OrbiterInclude
from orbiter.objects.pool import OrbiterPool
from orbiter.objects.requirement import OrbiterRequirement
from orbiter.objects.task import OrbiterOperator
from orbiter.objects.task_group import OrbiterTaskGroup
from orbiter.objects.timetables import OrbiterTimetable
from orbiter.objects.variable import OrbiterVariable

__mermaid__ = """
--8<-- [start:mermaid-relationships]
OrbiterProject --> "many" OrbiterConnection
OrbiterProject --> "many" OrbiterDAG
OrbiterProject --> "many" OrbiterEnvVar
OrbiterProject --> "many" OrbiterInclude
OrbiterProject --> "many" OrbiterPool
OrbiterProject --> "many" OrbiterRequirement
OrbiterProject --> "many" OrbiterVariable
--8<-- [end:mermaid-relationships]
"""


class OrbiterProject:
    # noinspection PyUnresolvedReferences
    """Holds everything necessary to render an Airflow Project.
    This is generated by a [`TranslationRuleset.translate_fn`][orbiter.rules.rulesets.TranslationRuleset].

    !!! tip

        They can be added together
        ```pycon
        >>> OrbiterProject() + OrbiterProject()
        OrbiterProject(dags=[], requirements=[], pools=[], connections=[], variables=[], env_vars=[])

        ```

        And compared
        ```
        >>> OrbiterProject() == OrbiterProject()
        True

        ```

    :param connections: A dictionary of [OrbiterConnections][orbiter.objects.connection.OrbiterConnection]
    :type connections: Dict[str, OrbiterConnection]
    :param dags: A dictionary of [OrbiterDAGs][orbiter.objects.dag.OrbiterDAG]
    :type dags: Dict[str, OrbiterDAG]
    :param env_vars: A dictionary of [OrbiterEnvVars][orbiter.objects.env_var.OrbiterEnvVar]
    :type env_vars: Dict[str, OrbiterEnvVar]
    :param includes: A dictionary of [OrbiterIncludes][orbiter.objects.include.OrbiterInclude]
    :type includes: Dict[str, OrbiterInclude]
    :param pools: A dictionary of [OrbiterPools][orbiter.objects.pool.OrbiterPool]
    :type pools: Dict[str, OrbiterPool]
    :param requirements: A set of [OrbiterRequirements][orbiter.objects.requirement.OrbiterRequirement]
    :type requirements: Set[OrbiterRequirement]
    :param variables: A dictionary of [OrbiterVariables][orbiter.objects.variable.OrbiterVariable]
    :type variables: Dict[str, OrbiterVariable]
    """

    # --8<-- [start:mermaid-props]
    connections: Dict[str, OrbiterConnection]
    dags: Dict[str, OrbiterDAG]
    env_vars: Dict[str, OrbiterEnvVar]
    includes: Dict[str, OrbiterInclude]
    pools: Dict[str, OrbiterPool]
    requirements: Set[OrbiterRequirement]
    variables: Dict[str, OrbiterVariable]
    # --8<-- [end:mermaid-props]

    def __init__(self):
        self.dags: Dict[str, OrbiterDAG] = dict()
        self.requirements: Set[OrbiterRequirement] = set()
        self.pools: Dict[str, OrbiterPool] = dict()
        self.connections: Dict[str, OrbiterConnection] = dict()
        self.variables: Dict[str, OrbiterVariable] = dict()
        self.env_vars: Dict[str, OrbiterEnvVar] = dict()
        self.includes: Dict[str, OrbiterInclude] = dict()

    def __add__(self, other) -> "OrbiterProject":
        if not other:
            return self
        if not isinstance(other, OrbiterProject):
            raise TypeError(f"Expected OrbiterProject, got {type(other)}")
        self.add_dags(other.dags.values())
        self.add_requirements(other.requirements)
        self.add_pools(other.pools.values())
        self.add_connections(other.connections.values())
        self.add_variables(other.variables.values())
        self.add_env_vars(other.env_vars.values())
        self.add_includes(other.includes.values())
        return self

    def __eq__(self, other) -> bool:
        return all(
            [str(self.dags[d]) == str(other.dags[d]) for d in self.dags]
            + [str(self.dags[d]) == str(other.dags[d]) for d in other.dags]
            + [self.requirements == other.requirements]
            + [self.pools == other.pools]
            + [self.connections == other.connections]
            + [self.variables == other.variables]
            + [self.env_vars == other.env_vars]
        )

    def __repr__(self):
        return (
            f"OrbiterProject("
            f"dags=[{','.join(sorted(self.dags.keys()))}], "
            f"requirements={sorted(self.requirements)}, "
            f"pools={sorted(self.pools)}, "
            f"connections={sorted(self.connections)}, "
            f"variables={sorted(self.variables)}, "
            f"env_vars={sorted(self.env_vars)})"
        )

    def add_connections(
        self, connections: OrbiterConnection | Iterable[OrbiterConnection]
    ) -> "OrbiterProject":
        """Add [`OrbiterConnections`][orbiter.objects.connection.OrbiterConnection] to the Project
        or override an existing connection with new properties

        ```pycon
        >>> OrbiterProject().add_connections(OrbiterConnection(conn_id='foo')).connections
        {'foo': OrbiterConnection(conn_id=foo, conn_type=generic)}

        >>> OrbiterProject().add_connections(
        ...     [OrbiterConnection(conn_id='foo'), OrbiterConnection(conn_id='bar')]
        ... ).connections
        {'foo': OrbiterConnection(conn_id=foo, conn_type=generic), 'bar': OrbiterConnection(conn_id=bar, conn_type=generic)}

        ```

        !!! tip

            Validation requires an `OrbiterConnection` to be passed
            ```pycon
            >>> # noinspection PyTypeChecker
            ... OrbiterProject().add_connections('foo')
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...
            >>> # noinspection PyTypeChecker
            >>> OrbiterProject().add_connections(['foo'])
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...

            ```

        :param connections: List of [`OrbiterConnections`][orbiter.objects.connection.OrbiterConnection]
        :type connections: List[OrbiterConnection] | OrbiterConnection
        :return: self
        :rtype: OrbiterProject
        """  # noqa: E501
        for connection in (
            [connections] if isinstance(connections, OrbiterConnection) else connections
        ):
            self.connections[connection.conn_id] = connection
        return self

    # noinspection t
    def add_dags(self, dags: OrbiterDAG | Iterable[OrbiterDAG]) -> "OrbiterProject":
        """Add [OrbiterDAGs][orbiter.objects.dag.OrbiterDAG]
        (and any [OrbiterRequirements][orbiter.objects.requirement.OrbiterRequirement],
        [OrbiterConns][orbiter.objects.connection.OrbiterConnection],
        [OrbiterVars][orbiter.objects.variable.OrbiterVariable],
        [OrbiterPools][orbiter.objects.pool.OrbiterPool],
        [OrbiterEnvVars][orbiter.objects.env_var.OrbiterEnvVar], etc.)
        to the Project.

        ```pycon
        >>> OrbiterProject().add_dags(OrbiterDAG(dag_id='foo', file_path="")).dags['foo'].repr()
        'OrbiterDAG(dag_id=foo, schedule=None, start_date=1970-01-01 00:00:00, catchup=False)'

        >>> dags = OrbiterProject().add_dags(
        ...     [OrbiterDAG(dag_id='foo', file_path=""), OrbiterDAG(dag_id='bar', file_path="")]
        ... ).dags; dags['foo'].repr(), dags['bar'].repr()
        ... # doctest: +NORMALIZE_WHITESPACE
        ('OrbiterDAG(dag_id=foo, schedule=None, start_date=1970-01-01 00:00:00, catchup=False)', 'OrbiterDAG(dag_id=bar, schedule=None, start_date=1970-01-01 00:00:00, catchup=False)')

        >>> # An example adding a little of everything, including deeply nested things
        ... from orbiter.objects.operators.bash import OrbiterBashOperator
        >>> from orbiter.objects.timetables.multi_cron_timetable import OrbiterMultiCronTimetable
        >>> from orbiter.objects.callbacks.smtp import OrbiterSmtpNotifierCallback
        >>> OrbiterProject().add_dags(OrbiterDAG(
        ...     dag_id='foo', file_path="",
        ...     orbiter_env_vars={OrbiterEnvVar(key="foo", value="bar")},
        ...     orbiter_includes={OrbiterInclude(filepath='foo.txt', contents="Hello, World!")},
        ...     schedule=OrbiterMultiCronTimetable(cron_defs=["0 */5 * * *", "0 */3 * * *"]),
        ...     tasks={'foo': OrbiterTaskGroup(task_group_id="foo",
        ...         tasks=[OrbiterBashOperator(
        ...             task_id='foo', bash_command='echo "Hello, World!"',
        ...             orbiter_pool=OrbiterPool(name='foo', slots=1),
        ...             orbiter_vars={OrbiterVariable(key='foo', value='bar')},
        ...             orbiter_conns={OrbiterConnection(conn_id='foo')},
        ...             orbiter_env_vars={OrbiterEnvVar(key='foo', value='bar')},
        ...             on_success_callback=OrbiterSmtpNotifierCallback(
        ...                 to="foo@bar.com",
        ...                 smtp_conn_id="SMTP",
        ...                 orbiter_conns={OrbiterConnection(conn_id="SMTP", conn_type="smtp")}
        ...             )
        ...         )]
        ...     )}
        ... ))
        ... # doctest: +NORMALIZE_WHITESPACE
        OrbiterProject(dags=[foo],
        requirements=[OrbiterRequirements(names=[DAG], package=apache-airflow, module=airflow, sys_package=None),
        OrbiterRequirements(names=[BashOperator], package=apache-airflow, module=airflow.operators.bash, sys_package=None),
        OrbiterRequirements(names=[send_smtp_notification], package=apache-airflow-providers-smtp, module=airflow.providers.smtp.notifications.smtp, sys_package=None),
        OrbiterRequirements(names=[TaskGroup], package=apache-airflow, module=airflow.utils.task_group, sys_package=None),
        OrbiterRequirements(names=[MultiCronTimetable], package=croniter, module=multi_cron_timetable, sys_package=None),
        OrbiterRequirements(names=[DateTime,Timezone], package=pendulum, module=pendulum, sys_package=None)],
        pools=['foo'],
        connections=['SMTP', 'foo'],
        variables=['foo'],
        env_vars=['foo'])

        ```

        !!! tip

            Validation requires an `OrbiterDAG` to be passed
            ```pycon
            >>> # noinspection PyTypeChecker
            ... OrbiterProject().add_dags('foo')
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...
            >>> # noinspection PyTypeChecker
            >>> OrbiterProject().add_dags(['foo'])
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...

            ```

        :param dags: List of [OrbiterDAGs][orbiter.objects.dag.OrbiterDAG]
        :type dags: List[OrbiterDAG] | OrbiterDAG
        :return: self
        :rtype: OrbiterProject
        """  # noqa: E501

        # noinspection t
        def _add_recursively(
            things: Iterable[
                OrbiterOperator
                | OrbiterTaskGroup
                | OrbiterCallback
                | OrbiterTimetable
                | OrbiterDAG
            ],
        ):
            for thing in things:
                if isinstance(thing, str):
                    continue
                if hasattr(thing, "orbiter_pool") and (pool := thing.orbiter_pool):
                    self.add_pools(pool)
                if hasattr(thing, "orbiter_conns") and (conns := thing.orbiter_conns):
                    self.add_connections(conns)
                if hasattr(thing, "orbiter_vars") and (variables := thing.orbiter_vars):
                    self.add_variables(variables)
                if hasattr(thing, "orbiter_env_vars") and (
                    env_vars := thing.orbiter_env_vars
                ):
                    self.add_env_vars(env_vars)
                if hasattr(thing, "orbiter_includes") and (
                    includes := thing.orbiter_includes
                ):
                    self.add_includes(includes)
                if hasattr(thing, "imports") and (imports := thing.imports):
                    self.add_requirements(imports)
                if isinstance(thing, OrbiterTaskGroup) and (tasks := thing.tasks):
                    _add_recursively(tasks)
                if hasattr(thing, "__dict__") or hasattr(thing, "model_extra"):
                    # If it's a pydantic model or dict, check its properties for more things to add
                    _add_recursively(
                        (
                            (getattr(thing, "__dict__", {}) or dict())
                            | (getattr(thing, "model_extra", {}) or dict())
                        ).values()
                    )

        for dag in [dags] if isinstance(dags, OrbiterDAG) else dags:
            dag_id = dag.dag_id

            # Add or update the DAG
            if dag_id in self.dags:
                self.dags[dag_id] += dag
            else:
                self.dags[dag_id] = dag

            # Add anything that might be in the tasks of the DAG - such as imports, Connections, etc
            _add_recursively((dag.tasks or {}).values())

            # Add anything that might be in the `dag.schedule` - such as Includes, Timetables, Connections, etc
            _add_recursively([dag])
        return self

    def add_env_vars(
        self, env_vars: OrbiterEnvVar | Iterable[OrbiterEnvVar]
    ) -> "OrbiterProject":
        """
        Add [OrbiterEnvVars][orbiter.objects.env_var.OrbiterEnvVar] to the Project
        or override an existing env var with new properties

        ```pycon
        >>> OrbiterProject().add_env_vars(OrbiterEnvVar(key="foo", value="bar")).env_vars
        {'foo': OrbiterEnvVar(key='foo', value='bar')}

        >>> OrbiterProject().add_env_vars([OrbiterEnvVar(key="foo", value="bar")]).env_vars
        {'foo': OrbiterEnvVar(key='foo', value='bar')}

        ```

        !!! tip

            Validation requires an `OrbiterEnvVar` to be passed
            ```pycon
            >>> # noinspection PyTypeChecker
            ... OrbiterProject().add_env_vars('foo')
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...

            >>> # noinspection PyTypeChecker
            ... OrbiterProject().add_env_vars(['foo'])
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...

            ```

        :param env_vars: List of [OrbiterEnvVar][orbiter.objects.env_var.OrbiterEnvVar]
        :type env_vars: List[OrbiterEnvVar] | OrbiterEnvVar
        :return: self
        :rtype: OrbiterProject
        """
        for env_var in [env_vars] if isinstance(env_vars, OrbiterEnvVar) else env_vars:
            self.env_vars[env_var.key] = env_var
        return self

    def add_includes(
        self, includes: OrbiterInclude | Iterable[OrbiterInclude]
    ) -> "OrbiterProject":
        """Add [OrbiterIncludes][orbiter.objects.include.OrbiterInclude] to the Project
        or override an existing [OrbiterInclude][orbiter.objects.include.OrbiterInclude] with new properties

        ```pycon
        >>> OrbiterProject().add_includes(OrbiterInclude(filepath="foo", contents="bar")).includes
        {'foo': OrbiterInclude(filepath='foo', contents='bar')}

        >>> OrbiterProject().add_includes([OrbiterInclude(filepath="foo", contents="bar")]).includes
        {'foo': OrbiterInclude(filepath='foo', contents='bar')}

        ```

        !!! tip

            Validation requires an `OrbiterInclude` to be passed
            ```pycon
            >>> # noinspection PyTypeChecker
            ... OrbiterProject().add_includes('foo')
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...

            >>> # noinspection PyTypeChecker
            ... OrbiterProject().add_includes(['foo'])
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...

            ```
        :param includes: List of [OrbiterIncludes][orbiter.objects.include.OrbiterInclude]
        :type includes: List[OrbiterInclude]
        :return: self
        :rtype: OrbiterProject
        """
        for include in [includes] if isinstance(includes, OrbiterInclude) else includes:
            self.includes[include.filepath] = include
        return self

    def add_pools(self, pools: OrbiterPool | Iterable[OrbiterPool]) -> "OrbiterProject":
        """Add [OrbiterPool][orbiter.objects.pool.OrbiterPool] to the Project
        or override existing pools with new properties

        ```pycon
        >>> OrbiterProject().add_pools(OrbiterPool(name="foo", slots=1)).pools
        {'foo': OrbiterPool(name='foo', description='', slots=1)}

        >>> ( OrbiterProject()
        ...     .add_pools([OrbiterPool(name="foo", slots=1)])
        ...     .add_pools([OrbiterPool(name="foo", slots=2)])
        ...     .pools
        ... )
        {'foo': OrbiterPool(name='foo', description='', slots=2)}

        ```

        !!! tip

            Validation requires an `OrbiterPool` to be passed
            ```pycon
            >>> # noinspection PyTypeChecker
            ... OrbiterProject().add_pools('foo')
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...
            >>> # noinspection PyTypeChecker
            ... OrbiterProject().add_pools(['foo'])
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...

            ```
        :param pools: List of [OrbiterPools][orbiter.objects.pool.OrbiterPool]
        :type pools: List[OrbiterPool] | OrbiterPool
        :return: self
        :rtype: OrbiterProject
        """
        for pool in [pools] if isinstance(pools, OrbiterPool) else pools:
            if pool.name in self.pools:
                self.pools[pool.name] += pool
            else:
                self.pools[pool.name] = pool
        return self

    def add_requirements(
        self, requirements: OrbiterRequirement | Iterable[OrbiterRequirement]
    ) -> "OrbiterProject":
        """Add [OrbiterRequirements][orbiter.objects.requirement.OrbiterRequirement] to the Project
        or override an existing requirement with new properties

        ```pycon
        >>> OrbiterProject().add_requirements(
        ...    OrbiterRequirement(package="apache-airflow", names=['foo'], module='bar'),
        ... ).requirements
        {OrbiterRequirements(names=[foo], package=apache-airflow, module=bar, sys_package=None)}

        >>> OrbiterProject().add_requirements(
        ...    [OrbiterRequirement(package="apache-airflow", names=['foo'], module='bar')],
        ... ).requirements
        {OrbiterRequirements(names=[foo], package=apache-airflow, module=bar, sys_package=None)}

        ```

        !!! tip

            Validation requires an `OrbiterRequirement` to be passed
            ```pycon
            >>> # noinspection PyTypeChecker
            ... OrbiterProject().add_requirements('foo')
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...
            >>> # noinspection PyTypeChecker
            >>> OrbiterProject().add_requirements(['foo'])
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...

            ```
        :param requirements: List of [OrbiterRequirements][orbiter.objects.requirement.OrbiterRequirement]
        :type requirements: List[OrbiterRequirement] | OrbiterRequirement
        :return: self
        :rtype: OrbiterProject
        """
        for requirement in (
            [requirements]
            if isinstance(requirements, OrbiterRequirement)
            else requirements
        ):
            self.requirements.add(requirement)
        return self

    def add_variables(
        self, variables: OrbiterVariable | Iterable[OrbiterVariable]
    ) -> "OrbiterProject":
        """Add [OrbiterVariables][orbiter.objects.variable.OrbiterVariable] to the Project
        or override an existing variable with new properties

        ```pycon
        >>> OrbiterProject().add_variables(OrbiterVariable(key="foo", value="bar")).variables
        {'foo': OrbiterVariable(key='foo', value='bar')}

        >>> OrbiterProject().add_variables([OrbiterVariable(key="foo", value="bar")]).variables
        {'foo': OrbiterVariable(key='foo', value='bar')}

        ```

        !!! tip

            Validation requires an `OrbiterVariable` to be passed
            ```pycon
            >>> # noinspection PyTypeChecker
            ... OrbiterProject().add_variables('foo')
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...
            >>> # noinspection PyTypeChecker
            ... OrbiterProject().add_variables(['foo'])
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...

            ```
        :param variables: List of [OrbiterVariable][orbiter.objects.variable.OrbiterVariable]
        :type variables: List[OrbiterVariable] | OrbiterVariable
        :return: self
        :rtype: OrbiterProject
        """
        for variable in (
            [variables] if isinstance(variables, OrbiterVariable) else variables
        ):
            self.variables[variable.key] = variable
        return self

    def render(self, output_dir: Path) -> None:
        if not len(self.dags):
            raise RuntimeError("No DAGs to render!")

        # /dags
        dags = output_dir / "dags"
        logger.info(f"Writing {dags}")
        dags.mkdir(parents=True, exist_ok=True)

        # /dags/*
        for dag_id, dag in self.dags.items():
            dag_file = dags / dag.file_path
            logger.debug(f"Writing {dag.file_path}")
            dag_file.parent.mkdir(parents=True, exist_ok=True)
            dag_file.write_text(str(dag))

        # requirements.txt
        if len([1 for r in self.requirements if r.package]):
            requirements = output_dir / "requirements.txt"
            logger.info(f"Writing {requirements}")
            requirements.write_text(
                "\n".join(
                    sorted(set(r.package for r in self.requirements if r.package))
                )
            )
        else:
            logger.debug("No python packages to write, skipping requirements.txt...")

        # packages.txt
        if len([1 for r in self.requirements if r.sys_package]):
            packages = output_dir / "packages.txt"
            logger.info(f"Writing {packages}")
            packages.write_text(
                "\n".join(
                    sorted(
                        set(r.sys_package for r in self.requirements if r.sys_package)
                    )
                )
            )
        else:
            logger.debug("No system packages to write, skipping packages.txt...")

        # Pools, Connections, Variables
        if any((len(self.pools), len(self.variables), len(self.connections))):
            airflow_settings = output_dir / "airflow_settings.yaml"
            logger.info(f"Writing {airflow_settings}")
            yaml.safe_dump(
                {
                    "airflow": {
                        "pools": [pool.render() for pool in self.pools.values()],
                        "variables": [
                            variable.render() for variable in self.variables.values()
                        ],
                        "connections": [
                            connection.render()
                            for connection in self.connections.values()
                        ],
                    }
                },
                airflow_settings.open("w"),
            )
        else:
            logger.debug(
                "No Pools, Variables, or Connections to write. Skipping airflow_settings.yaml..."
            )

        # /include
        if len(self.includes):
            for include in self.includes.values():
                include_path = output_dir / include.filepath
                logger.info(f"Writing {include_path}")
                include_path.write_text(include.render())
        else:
            logger.debug("No files to include")

        # ENV VAR
        if len(self.env_vars):
            dotenv = output_dir / ".env"
            logger.info(f"Writing {dotenv}")
            dotenv.write_text("\n".join(env.render() for env in self.env_vars.values()))
        else:
            logger.debug("No entries for .env")

    @validate_call
    def analyze(
        self, output_fmt: Literal["json", "csv", "md"] = "md", output_file=None
    ):
        """Print an analysis of the project to the console.

        !!! tip

            Looks for a specific `[task_type=XYZ]` in the Task's `doc_md` property
            or uses `type(task)` to infer the type of task.

        ```pycon
        >>> from orbiter.objects.operators.empty import OrbiterEmptyOperator
        >>> OrbiterProject().add_dags([
        ...     OrbiterDAG(file_path="", dag_id="foo", orbiter_kwargs={"file_path": "foo.py"},
        ...         tasks={"bar": OrbiterEmptyOperator(task_id="bar")}
        ...     ),
        ...     OrbiterDAG(file_path="", dag_id="baz", orbiter_kwargs={"file_path": "baz.py"},
        ...         tasks={"bop": OrbiterEmptyOperator(task_id="bop")}
        ...     )
        ... ]).analyze()
        ... # doctest: +ELLIPSIS
        ┏━...
        ...Analysis...
        ┗━...
        <BLANKLINE>
        <BLANKLINE>
                   DAGs   OrbiterEmptyOperator
         ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
          foo.py      1                      1
          baz.py      1                      1
          Totals      2                      2
        <BLANKLINE>

        ```
        """
        if output_file is None:
            output_file = sys.stdout

        _task_type = re.compile(r"\[task_type=(?P<task_type>[A-Za-z0-9-_]+)")

        def get_task_type(task):
            match = _task_type.match(getattr(task, "doc_md", None) or "")
            match_or_task_type = (
                match.groupdict().get("task_type") if match else None
            ) or type(task).__name__
            return match_or_task_type

        dag_analysis = [
            {
                "file": dag.orbiter_kwargs.get("file_path", dag.file_path),
                "dag_id": dag.dag_id,
                "task_types": [get_task_type(task) for task in dag.tasks.values()],
            }
            for dag in self.dags.values()
        ]

        file_analysis = {}
        for analysis in dag_analysis:
            analysis_output = file_analysis.get(analysis["file"], {})
            analysis_output["DAGs"] = analysis_output.get("DAGs", 0) + 1
            tasks_of_type = reduce(
                lambda acc, task_type: acc | {task_type: acc.get(task_type, 0) + 1},
                analysis["task_types"],
                dict(),
            )
            analysis_output |= tasks_of_type
            file_analysis[analysis["file"]] = analysis_output

        file_analysis = [{"": k} | v for k, v in file_analysis.items()]
        totals = {"": "Totals"}
        for file in file_analysis:
            for k, v in file.items():
                if k != "":
                    totals[k] = totals.get(k, 0) + v
        file_analysis.append(totals)

        if output_fmt == "json":
            import json

            json.dump(file_analysis, output_file, default=str)
        elif output_fmt == "csv":
            import csv

            writer = csv.DictWriter(output_file, fieldnames={""} | totals.keys())
            writer.writeheader()
            writer.writerows(file_analysis)
        elif output_fmt == "md":
            from rich.console import Console
            from rich.markdown import Markdown
            from tabulate import tabulate

            console = Console(file=output_file)

            #         DAGs  EmptyOp
            # file_a     1        1
            table = tabulate(
                tabular_data=file_analysis,
                headers="keys",
                tablefmt="pipe",
                # https://github.com/Textualize/rich/issues/3027
                missingval="⠀",  # (special 'braille space' character)
            )
            console.print(
                Markdown(
                    f"# Analysis\n{table}",
                    style="magenta",
                )
            )


# https://github.com/pydantic/pydantic/issues/8790
OrbiterProject.add_connections = validate_call()(OrbiterProject.add_connections)
OrbiterProject.add_dags = validate_call()(OrbiterProject.add_dags)
OrbiterProject.add_env_vars = validate_call()(OrbiterProject.add_env_vars)
OrbiterProject.add_includes = validate_call()(OrbiterProject.add_includes)
OrbiterProject.add_pools = validate_call()(OrbiterProject.add_pools)
OrbiterProject.add_requirements = validate_call()(OrbiterProject.add_requirements)
OrbiterProject.add_variables = validate_call()(OrbiterProject.add_variables)


if __name__ == "__main__":
    import doctest

    doctest.testmod(
        optionflags=doctest.ELLIPSIS
        | doctest.NORMALIZE_WHITESPACE
        | doctest.IGNORE_EXCEPTION_DETAIL
    )



from __future__ import annotations

from abc import ABC
from typing import ClassVar


class AirflowSettingsRender(ABC):
    """Used to render objects as `yaml` to `airflow_settings.yaml` file"""

    render_prefix: ClassVar[str]

    def render(self) -> dict:
        # noinspection PyUnresolvedReferences
        return {
            (self.render_prefix if not k.startswith(self.render_prefix) else "") + k: v
            for k, v in self.model_dump().items()
        }



from __future__ import annotations

from typing import ClassVar

from pydantic import BaseModel

from orbiter.objects.airflow_settings import AirflowSettingsRender


class OrbiterConnection(BaseModel, AirflowSettingsRender, extra="allow"):
    """An Airflow
    [Connection](https://airflow.apache.org/docs/apache-airflow-providers/core-extensions/connections.html),
    rendered to an
    [`airflow_settings.yaml`](https://www.astronomer.io/docs/astro/cli/develop-project#configure-airflow_settingsyaml-local-development-only)
    file.

    See also other
    [Connection](https://airflow.apache.org/docs/apache-airflow/stable/authoring-and-scheduling/connections.html)
    documentation.

    ```pycon
    >>> OrbiterConnection(
    ...     conn_id="my_conn_id", conn_type="mysql", host="localhost", port=3306, login="root"
    ... ).render()
    {'conn_id': 'my_conn_id', 'conn_type': 'mysql', 'conn_host': 'localhost', 'conn_port': 3306, 'conn_login': 'root'}

    ```

    !!! note

        Use the utility `conn_id` function to generate both an `OrbiterConnection`
        and connection property for an operator

        ```python
        from orbiter.objects import conn_id

        OrbiterTask(
            ... ,
            **conn_id("my_conn_id", conn_type="mysql"),
        )
        ```

    :param conn_id: The ID of the connection
    :type conn_id: str
    :param conn_type: The type of the connection, always lowercase. Defaults to 'generic'
    :type conn_type: str, optional
    :param **kwargs: Additional properties for the connection
    """  # noqa: E501

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    conn_id: str
    conn_type: str
    **kwargs
    --8<-- [end:mermaid-props]
    """
    render_prefix: ClassVar[str] = "conn_"

    conn_id: str
    conn_type: str = "generic"

    def __hash__(self):
        return hash(f"{self.model_dump_json()}")

    def __repr__(self):
        return (
            "OrbiterConnection("
            + ", ".join(f"{k}={v}" for k, v in self.model_dump().items())
            + ")"
        )



from __future__ import annotations

from typing import Dict

from orbiter.objects import ImportList
from orbiter.objects.requirement import OrbiterRequirement
from orbiter.objects.task import OrbiterOperator, RenderAttributes

__mermaid__ = """
--8<-- [start:mermaid-relationships]
OrbiterOperator "implements" <|-- OrbiterSSHOperator
--8<-- [end:mermaid-relationships]
"""


class OrbiterSSHOperator(OrbiterOperator):
    """
    An Airflow
    [SSHOperator](https://registry.astronomer.io/providers/apache-airflow-providers-ssh/versions/latest/modules/SSHOperator).
    Used to run shell commands over SSH.

    ```pycon
    >>> OrbiterSSHOperator(task_id="foo", ssh_conn_id="SSH", command="echo 'hello world'")
    foo_task = SSHOperator(task_id='foo', ssh_conn_id='SSH', command="echo 'hello world'")

    ```
    :param task_id: The `task_id` for the operator
    :type task_id: str
    :param ssh_conn_id: The SSH connection to use. (Note: use the `**conn_id(...)` utility function)
    :type ssh_conn_id: str
    :param command: The command to execute
    :type command: str
    :param environment: The environment variables to set, defaults to None
    :type environment: dict, optional
    :param **kwargs: Extra arguments to pass to the operator
    :param **OrbiterBase: [OrbiterBase][orbiter.objects.OrbiterBase] inherited properties
    """  # noqa: E501

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    operator = "SSHOperator"
    task_id: str
    ssh_conn_id: str
    command: str
    environment: Dict[str, str] | None
    --8<-- [end:mermaid-props]
    """

    imports: ImportList = [
        OrbiterRequirement(
            package="apache-airflow-providers-ssh",
            module="airflow.providers.ssh.operators.ssh",
            names=["SSHOperator"],
        )
    ]
    operator: str = "SSHOperator"
    # noinspection Pydantic
    render_attributes: RenderAttributes = OrbiterOperator.render_attributes + [
        "ssh_conn_id",
        "command",
        "environment",
    ]

    ssh_conn_id: str
    command: str
    environment: Dict[str, str] | None = None



from __future__ import annotations
from typing import Callable

from orbiter.ast_helper import py_function
from orbiter.objects import ImportList
from orbiter.objects.requirement import OrbiterRequirement
from orbiter.objects.task import OrbiterOperator, RenderAttributes

__mermaid__ = """
--8<-- [start:mermaid-relationships]
OrbiterOperator "implements" <|-- OrbiterPythonOperator
--8<-- [end:mermaid-relationships]
"""


class OrbiterPythonOperator(OrbiterOperator):
    """
    An Airflow
    [PythonOperator](https://registry.astronomer.io/providers/apache-airflow/versions/latest/modules/PythonOperator).
    Used to execute any Python Function.

    ```pycon
    >>> def foo(a, b):
    ...    print(a + b)
    >>> OrbiterPythonOperator(task_id="foo", python_callable=foo)
    def foo(a, b):
       print(a + b)
    foo_task = PythonOperator(task_id='foo', python_callable=foo)

    ```
    :param task_id: The `task_id` for the operator
    :type task_id: str
    :param python_callable: The python function to execute
    :type python_callable: Callable
    :param op_args: The arguments to pass to the python function, defaults to None
    :type op_args: list | None, optional
    :param op_kwargs: The keyword arguments to pass to the python function, defaults to None
    :type op_kwargs: dict | None, optional
    :param **kwargs: Extra arguments to pass to the operator
    :param **OrbiterBase: [OrbiterBase][orbiter.objects.OrbiterBase] inherited properties
    """

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    operator = "PythonOperator"
    task_id: str
    python_callable: Callable
    op_args: list | None
    op_kwargs: dict | None
    --8<-- [end:mermaid-props]
    """

    imports: ImportList = [
        OrbiterRequirement(
            package="apache-airflow",
            module="airflow.operators.python",
            names=["PythonOperator"],
        )
    ]
    operator: str = "PythonOperator"
    # noinspection Pydantic
    render_attributes: RenderAttributes = OrbiterOperator.render_attributes + [
        "python_callable",
        "op_args",
        "op_kwargs",
    ]

    python_callable: Callable
    op_args: list | None = None
    op_kwargs: dict | None = None

    def _to_ast(self):
        return [py_function(self.python_callable), super()._to_ast()]



from __future__ import annotations

import ast
from abc import ABC
from typing import List, Any, Set

from pydantic import field_validator

from orbiter.config import ORBITER_TASK_SUFFIX
from orbiter.ast_helper import (
    OrbiterASTBase,
    py_with,
    py_object,
    py_bitshift,
)
from orbiter.objects import OrbiterBase, ImportList, OrbiterRequirement
from orbiter.objects.task import (
    TaskId,
    OrbiterTaskDependency,
    OrbiterOperator,
    task_add_downstream,
    to_task_id,
)

__mermaid__ = """
--8<-- [start:mermaid-dag-relationships]
OrbiterTaskGroup --> "many" OrbiterRequirement
--8<-- [end:mermaid-dag-relationships]

--8<-- [start:mermaid-task-relationships]
--8<-- [end:mermaid-task-relationships]
"""


class OrbiterTaskGroup(OrbiterASTBase, OrbiterBase, ABC, extra="forbid"):
    """
    Represents a [TaskGroup](https://airflow.apache.org/docs/apache-airflow/stable/core-concepts/dags.html#taskgroups)
    in Airflow, which contains multiple tasks

    ```pycon
    >>> from orbiter.objects.operators.bash import OrbiterBashOperator
    >>> from orbiter.ast_helper import render_ast
    >>> OrbiterTaskGroup(task_group_id="foo", tasks=[
    ...   OrbiterBashOperator(task_id="b", bash_command="b"),
    ...   OrbiterBashOperator(task_id="a", bash_command="a").add_downstream("b"),
    ... ], downstream={"c"})
    with TaskGroup(group_id='foo') as foo:
        b_task = BashOperator(task_id='b', bash_command='b')
        a_task = BashOperator(task_id='a', bash_command='a')
        a_task >> b_task

    >>> render_ast(OrbiterTaskGroup(task_group_id="foo", tasks=[], downstream={"c"})._downstream_to_ast())
    'foo >> c_task'

    ```

    :param task_group_id: The id of the TaskGroup
    :type task_group_id: str
    :param tasks: The tasks in the TaskGroup
    :type tasks: List[OrbiterOperator | OrbiterTaskGroup]
    :param **OrbiterBase: [OrbiterBase][orbiter.objects.OrbiterBase] inherited properties
    """

    __mermaid__ = """
       --8<-- [start:mermaid-op-props]
    task_group_id: str
    tasks: List[OrbiterOperator | OrbiterTaskGroup]
    add_downstream(str | List[str] | OrbiterTaskDependency)
    --8<-- [end:mermaid-op-props]
    """

    imports: ImportList = [
        OrbiterRequirement(
            package="apache-airflow",
            module="airflow.utils.task_group",
            names=["TaskGroup"],
        )
    ]
    task_group_id: TaskId
    tasks: List[Any]
    downstream: Set[str] = set()

    @property
    def task_id(self):
        # task_id property, so it can be treated like an OrbiterOperator more easily
        return self.task_group_id

    @task_id.setter
    def task_id(self, value):
        # task_id property, so it can be treated like an OrbiterOperator more easily
        self.task_group_id = value

    # noinspection PyNestedDecorators
    @field_validator("tasks")
    @classmethod
    def ensure_tasks(cls, tasks: Any):
        if any(
            not (
                issubclass(type(i), OrbiterOperator) or isinstance(i, OrbiterTaskGroup)
            )
            for i in tasks
        ):
            raise TypeError(
                f"At least one item in tasks={tasks} is not a subclass of OrbiterOperator nor an OrbiterTaskGroup"
            )
        return tasks

    def add_downstream(
        self, task_id: str | List[str] | OrbiterTaskDependency
    ) -> "OrbiterTaskGroup":
        return task_add_downstream(self, task_id)

    def _downstream_to_ast(self):
        if not self.downstream:
            return
        elif len(self.downstream) == 1:
            (t,) = tuple(self.downstream)
            return py_bitshift(
                to_task_id(self.task_id), to_task_id(t, ORBITER_TASK_SUFFIX)
            )
        else:
            return py_bitshift(
                to_task_id(self.task_id),
                sorted([to_task_id(t, ORBITER_TASK_SUFFIX) for t in self.downstream]),
            )

    def _to_ast(self) -> ast.stmt:
        # noinspection PyProtectedMember
        return py_with(
            py_object("TaskGroup", group_id=self.task_group_id).value,
            [operator._to_ast() for operator in self.tasks]
            + [operator._downstream_to_ast() for operator in self.tasks],
            self.task_group_id,
        )



from __future__ import annotations

import ast
from abc import ABC
from typing import Set, List, ClassVar, Annotated, Callable

from loguru import logger
from pydantic import AfterValidator, BaseModel, validate_call

from orbiter import clean_value
from orbiter.ast_helper import (
    OrbiterASTBase,
    py_function,
    py_bitshift,
)
from orbiter.ast_helper import py_assigned_object
from orbiter.config import ORBITER_TASK_SUFFIX
from orbiter.objects import ImportList
from orbiter.objects import OrbiterBase
from orbiter.objects.pool import OrbiterPool

__mermaid__ = """
--8<-- [start:mermaid-dag-relationships]
OrbiterOperator --> "many" OrbiterRequirement
OrbiterOperator --> "one" OrbiterPool
OrbiterOperator --> "many" OrbiterConnection
OrbiterOperator --> "many" OrbiterVariable
OrbiterOperator --> "many" OrbiterEnvVar
--8<-- [end:mermaid-dag-relationships]

--8<-- [start:mermaid-task-relationships]
OrbiterOperator "implements" <|-- OrbiterTask
OrbiterOperator --> "many" OrbiterCallback
--8<-- [end:mermaid-task-relationships]
"""

RenderAttributes = ClassVar[List[str]]
TaskId = Annotated[str, AfterValidator(lambda t: to_task_id(t))]


def task_add_downstream(
    self, task_id: str | List[str] | OrbiterTaskDependency
) -> "OrbiterOperator" | "OrbiterTaskGroup":  # noqa: F821
    # noinspection PyProtectedMember
    """
    Add a downstream task dependency

    ```pycon
    >>> from orbiter.objects.operators.empty import OrbiterEmptyOperator
    >>> from orbiter.objects.task import task_add_downstream, OrbiterTaskDependency
    >>> from orbiter.ast_helper import render_ast
    >>> render_ast(task_add_downstream(OrbiterEmptyOperator(task_id="task_id"), "downstream")._downstream_to_ast())
    'task_id_task >> downstream_task'
    >>> render_ast(
    ...     task_add_downstream(OrbiterEmptyOperator(task_id="task_id"),
    ...     OrbiterTaskDependency(task_id="task_id", downstream="downstream"))._downstream_to_ast()
    ... )
    'task_id_task >> downstream_task'
    >>> render_ast(task_add_downstream(
    ...     OrbiterEmptyOperator(task_id="task_id"),
    ...     ["downstream"]
    ... )._downstream_to_ast())
    'task_id_task >> downstream_task'
    >>> render_ast(task_add_downstream(
    ...     OrbiterEmptyOperator(task_id="task_id"),
    ...     ["downstream", "downstream2"]
    ... )._downstream_to_ast())
    'task_id_task >> [downstream2_task, downstream_task]'

    ```
    """  # noqa: E501
    if isinstance(task_id, OrbiterTaskDependency):
        task_dependency = task_id
        if task_dependency.task_id != self.task_id:
            raise ValueError(
                f"task_dependency={task_dependency} has a different task_id than {self.task_id}"
            )
        # do normal parsing logic, but with these downstream items
        task_id = task_dependency.downstream

    if isinstance(task_id, str):
        self.downstream |= {to_task_id(task_id)}
        return self
    else:
        if not len(task_id):
            return self

        downstream_task_id = {to_task_id(t) for t in task_id}
        logger.debug(f"Adding downstream {downstream_task_id} to {self.task_id}")
        self.downstream |= downstream_task_id
        return self


class OrbiterTaskDependency(BaseModel, extra="forbid"):
    """Represents a task dependency, which is added to either an
    [`OrbiterOperator`][orbiter.objects.task.OrbiterOperator]
    or an [`OrbiterTaskGroup`][orbiter.objects.task_group.OrbiterTaskGroup].

    :param task_id: The task_id for the operator
    :type task_id: str
    :param downstream: downstream task(s)
    :type downstream: str | List[str]
    """

    # --8<-- [start:mermaid-td-props]
    task_id: TaskId
    downstream: TaskId | List[TaskId]
    # --8<-- [end:mermaid-td-props]

    def __str__(self):
        return f"{self.task_id} >> {self.downstream}"

    def __repr__(self):
        return str(self)


class OrbiterOperator(OrbiterASTBase, OrbiterBase, ABC, extra="allow"):
    """
    **Abstract class** representing a
    [Task in Airflow](https://airflow.apache.org/docs/apache-airflow/stable/tutorial/fundamentals.html#operators).

    **Must be subclassed** (such as [`OrbiterBashOperator`][orbiter.objects.operators.bash.OrbiterBashOperator],
    or [`OrbiterTask`][orbiter.objects.task.OrbiterTask]).

    Subclassing Example:
    ```pycon
    >>> from orbiter.objects import OrbiterRequirement
    >>> class OrbiterMyOperator(OrbiterOperator):
    ...   imports: ImportList = [OrbiterRequirement(package="apache-airflow")]
    ...   operator: str = "MyOperator"

    >>> foo = OrbiterMyOperator(task_id="task_id"); foo
    task_id_task = MyOperator(task_id='task_id')

    ```

    Adding single downstream tasks:
    ```pycon
    >>> from orbiter.ast_helper import render_ast
    >>> render_ast(foo.add_downstream("downstream")._downstream_to_ast())
    'task_id_task >> downstream_task'

    ```

    Adding multiple downstream tasks:
    ```pycon
    >>> render_ast(foo.add_downstream(["a", "b"])._downstream_to_ast())
    'task_id_task >> [a_task, b_task, downstream_task]'

    ```

    !!! note

        Validation - task_id in OrbiterTaskDependency must match this task_id
        ```pycon
        >>> foo.add_downstream(OrbiterTaskDependency(task_id="other", downstream="bar")).downstream
        ... # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ValueError: Task dependency ... has a different task_id than task_id

        ```

    :param imports: List of requirements for the operator
    :type imports: List[OrbiterRequirement]
    :param task_id: The `task_id` for the operator, must be unique and snake_case
    :type task_id: str
    :param trigger_rule: Conditions under which to start the task
        [(docs)](https://airflow.apache.org/docs/apache-airflow/stable/concepts.html#trigger-rules)
    :type trigger_rule: str, optional
    :param pool: Name of the
        [pool](https://airflow.apache.org/docs/apache-airflow/stable/administration-and-deployment/pools.html#pools)
        to use
    :type pool: str, optional
    :param pool_slots: Slots for this task to occupy
    :type pool_slots: int, optional
    :param operator: Operator name
    :type operator: str, optional
    :param downstream: Downstream tasks, defaults to `set()`
    :type downstream: Set[str], optional
    :param **kwargs: Other properties that may be passed to operator
    :param **OrbiterBase: [OrbiterBase][orbiter.objects.OrbiterBase] inherited properties
    """

    imports: ImportList

    operator: str
    task_id: TaskId

    trigger_rule: str | None = None

    pool: str | None = None
    pool_slots: int | None = None
    orbiter_pool: OrbiterPool | None = None
    downstream: Set[str] = set()

    render_attributes: RenderAttributes = [
        "task_id",
        "pool",
        "pool_slots",
        "trigger_rule",
    ]

    __mermaid__ = """
    --8<-- [start:mermaid-op-props]
    imports: List[OrbiterRequirement]
    operator: str
    task_id: str
    pool: str | None
    pool_slots: int | None
    trigger_rule: str | None
    downstream: Set[str]
    add_downstream(str | List[str] | OrbiterTaskDependency)
    --8<-- [end:mermaid-op-props]
    """

    def add_downstream(
        self, task_id: str | List[str] | OrbiterTaskDependency
    ) -> "OrbiterOperator":
        return task_add_downstream(self, task_id)

    def _downstream_to_ast(self) -> List[ast.stmt]:
        if not self.downstream:
            return []
        elif len(self.downstream) == 1:
            (t,) = tuple(self.downstream)
            return py_bitshift(
                to_task_id(self.task_id, ORBITER_TASK_SUFFIX),
                to_task_id(t, ORBITER_TASK_SUFFIX),
            )
        else:
            return py_bitshift(
                to_task_id(self.task_id, ORBITER_TASK_SUFFIX),
                sorted([to_task_id(t, ORBITER_TASK_SUFFIX) for t in self.downstream]),
            )

    def _to_ast(self) -> ast.stmt:
        def prop(k):
            attr = getattr(self, k, None) or getattr(self.model_extra, k, None)
            return ast.Name(id=attr.__name__) if isinstance(attr, Callable) else attr

        # foo = Bar(x=x,y=y, z=z)
        return py_assigned_object(
            to_task_id(self.task_id, ORBITER_TASK_SUFFIX),
            self.operator,
            **{k: prop(k) for k in self.render_attributes if k and getattr(self, k)},
            **{k: prop(k) for k in (self.model_extra.keys() or [])},
        )


class OrbiterTask(OrbiterOperator, extra="allow"):
    """
    A generic version of [`OrbiterOperator`][orbiter.objects.task.OrbiterOperator] that can be instantiated directly.

    The operator that is instantiated is inferred from the `imports` field,
    via the first `*Operator` or `*Sensor` import.

    [View info for specific operators at the Astronomer Registry.](https://registry.astronomer.io/)

    ```pycon
    >>> from orbiter.objects.requirement import OrbiterRequirement
    >>> OrbiterTask(task_id="foo", bash_command="echo 'hello world'", other=1, imports=[
    ...   OrbiterRequirement(package="apache-airflow", module="airflow.operators.bash", names=["BashOperator"])
    ... ])
    foo_task = BashOperator(task_id='foo', bash_command="echo 'hello world'", other=1)

    >>> def foo():
    ...   pass
    >>> OrbiterTask(task_id="foo", python_callable=foo, other=1, imports=[
    ...   OrbiterRequirement(package="apache-airflow", module="airflow.sensors.python", names=["PythonSensor"])
    ... ])
    def foo():
        pass
    foo_task = PythonSensor(task_id='foo', other=1, python_callable=foo)

    ```

    :param task_id: The `task_id` for the operator. Must be unique and snake_case
    :type task_id: str
    :param imports: List of requirements for the operator.
        The Operator is inferred from first `*Operator` or `*Sensor` imported.
    :type imports: List[OrbiterRequirement]
    :param **kwargs: Any other keyword arguments to be passed to the operator
    """

    imports: ImportList
    task_id: TaskId
    operator: None = None  # Not used

    __mermaid__ = """
    --8<-- [start:mermaid-task-props]
    <<OrbiterOperator>>
    imports: List[OrbiterRequirement]
    task_id: str
    **kwargs
    --8<-- [end:mermaid-task-props]
    """

    def _to_ast(self) -> ast.stmt:
        def prop(k):
            attr = getattr(self, k)
            return ast.Name(id=attr.__name__) if isinstance(attr, Callable) else attr

        # Figure out which operator we are talking about
        operator_names = [
            name
            for _import in self.imports
            for name in _import.names
            if "operator" in name.lower() or "sensor" in name.lower()
        ]
        if len(operator_names) != 1:
            raise ValueError(
                f"Expected exactly one operator name, got {operator_names}"
            )
        [operator] = operator_names

        self_as_ast = py_assigned_object(
            to_task_id(self.task_id, ORBITER_TASK_SUFFIX),
            operator,
            **{
                k: prop(k)
                for k in ["task_id"] + sorted(self.__pydantic_extra__.keys())
                if k and getattr(self, k)
            },
        )
        callable_props = [
            k
            for k in self.__pydantic_extra__.keys()
            if isinstance(getattr(self, k), Callable)
        ]
        return (
            [py_function(getattr(self, prop)) for prop in callable_props]
            + [self_as_ast]
            if len(callable_props)
            else self_as_ast
        )


@validate_call
def to_task_id(task_id: str, assignment_suffix: str = "") -> str:
    # noinspection PyTypeChecker
    """General utiltty function - turns MyTaskId into my_task_id (or my_task_id_task suffix is `_task`)
    ```pycon
    >>> to_task_id("MyTaskId")
    'my_task_id'
    >>> to_task_id("MyTaskId", "_task")
    'my_task_id_task'
    >>> to_task_id("my_task_id_task", "_task")
    'my_task_id_task'

    ```
    :param task_id:
    :type task_id: str
    :param assignment_suffix: e.g. `_task` for `task_id_task = MyOperator(...)`
    :type assignment_suffix: str
    """
    task_id = clean_value(task_id)
    return task_id + (
        assignment_suffix
        if task_id[-len(assignment_suffix) :] != assignment_suffix
        else ""
    )



from __future__ import annotations

import ast
from datetime import datetime
from functools import reduce
from pathlib import Path
from typing import Annotated, Any, Dict, Iterable, List, Callable

from pendulum import DateTime
from pydantic import AfterValidator, validate_call

from orbiter import clean_value
from orbiter.ast_helper import OrbiterASTBase, py_object, py_with
from orbiter.objects import ImportList, OrbiterBase, CALLBACK_KEYS
from orbiter.objects.requirement import OrbiterRequirement
from orbiter.objects.task import OrbiterOperator
from orbiter.objects.task_group import OrbiterTaskGroup
from orbiter.objects.timetables import OrbiterTimetable

__mermaid__ = """
--8<-- [start:mermaid-project-relationships]
OrbiterDAG --> "many" OrbiterInclude
OrbiterDAG --> "many" OrbiterConnection
OrbiterDAG --> "many" OrbiterEnvVar
OrbiterDAG --> "many" OrbiterRequirement
OrbiterDAG --> "many" OrbiterVariable
--8<-- [end:mermaid-project-relationships]

--8<-- [start:mermaid-dag-relationships]
OrbiterDAG "via schedule" --> OrbiterTimetable
OrbiterDAG --> "many" OrbiterOperator
OrbiterDAG --> "many" OrbiterTaskGroup
OrbiterDAG --> "many" OrbiterRequirement
--8<-- [end:mermaid-dag-relationships]
"""

DagId = Annotated[str, AfterValidator(lambda d: to_dag_id(d))]


def _get_imports_recursively(
    tasks: Iterable[OrbiterOperator | OrbiterTaskGroup],
) -> List[OrbiterRequirement]:
    """

    >>> from orbiter.objects.task import OrbiterTask
    >>> from orbiter.objects.task_group import OrbiterTaskGroup
    >>> from orbiter.objects.callbacks import OrbiterCallback
    >>> _get_imports_recursively([
    ...     OrbiterTask(task_id="foo", imports=[OrbiterRequirement(names=['foo'])]),
    ...     OrbiterTaskGroup(task_group_id="bar", imports=[OrbiterRequirement(names=['bar'])], tasks=[
    ...         OrbiterTask(task_id="baz", imports=[OrbiterRequirement(names=['baz'])],
    ...             on_failure_callback=OrbiterCallback(imports=[OrbiterRequirement(names=['qux'])], function='qux')
    ...        )
    ...     ])
    ... ])
    ... # doctest: +ELLIPSIS
    [OrbiterRequirements(...names=[bar]...names=[baz]...names=[foo]...names=[qux]...]

    """
    imports = set()
    for task in tasks:
        # Add task imports
        imports |= set(task.imports)

        def reduce_imports_from_callback(old, item):
            try:
                # Look for on_failure_callback
                task_props = (getattr(task, "model_extra", {}) or {}) | (
                    getattr(task, "__dict__", {}) or {}
                )
                callback = task_props.get(item)
                # get imports from callback, merge them all
                return old | set(getattr(callback, "imports"))
            except (AttributeError, KeyError):
                return old

        imports |= reduce(reduce_imports_from_callback, CALLBACK_KEYS, set())
        if hasattr(task, "tasks"):
            # descend, for a task group
            imports |= set(_get_imports_recursively(task.tasks))
    return list(sorted(imports, key=str))


class OrbiterDAG(OrbiterASTBase, OrbiterBase, extra="allow"):
    # noinspection PyUnresolvedReferences
    """Represents an Airflow [DAG](https://airflow.apache.org/docs/apache-airflow/stable/core-concepts/dags.html),
    with its tasks and dependencies.

    Renders to a `.py` file in the `/dags` folder

    :param file_path: File path of the DAG, relative to the `/dags` folder
        (`filepath=my_dag.py` would render to `dags/my_dag.py`)
    :type file_path: str
    :param dag_id: The `dag_id`. Must be unique and snake_case. Good practice is to set `dag_id` == `file_path`
    :type dag_id: str
    :param schedule: The schedule for the DAG. Defaults to None (only runs when manually triggered)
    :type schedule: str | OrbiterTimetable, optional
    :param catchup: Whether to catchup runs from the `start_date` to now, on first run. Defaults to False
    :type catchup: bool, optional
    :param start_date: The start date for the DAG. Defaults to Unix Epoch
    :type start_date: DateTime, optional
    :param tags: Tags for the DAG, used for sorting and filtering in the Airflow UI
    :type tags: List[str], optional
    :param default_args: Default arguments for any tasks in the DAG
    :type default_args: Dict[str, Any], optional
    :param params: Params for the DAG
    :type params: Dict[str, Any], optional
    :param doc_md: Documentation for the DAG with markdown support
    :type doc_md: str, optional
    :param kwargs: Additional keyword arguments to pass to the DAG
    :type kwargs: dict, optional
    :param **OrbiterBase: [OrbiterBase][orbiter.objects.OrbiterBase] inherited properties
    """

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    imports: List[OrbiterRequirement]
    file_path: str
    dag_id: str
    schedule: str | OrbiterTimetable | None
    catchup: bool
    start_date: DateTime
    tags: List[str]
    default_args: Dict[str, Any]
    params: Dict[str, Any]
    doc_md: str | None
    tasks: Dict[str, OrbiterOperator]
    kwargs: dict
    orbiter_kwargs: dict
    orbiter_conns: Set[OrbiterConnection]
    orbiter_vars: Set[OrbiterVariable]
    orbiter_env_vars: Set[OrbiterEnvVar]
    orbiter_includes: Set[OrbiterInclude]
    --8<-- [end:mermaid-props]
    """

    imports: ImportList = [
        OrbiterRequirement(package="apache-airflow", module="airflow", names=["DAG"]),
        OrbiterRequirement(
            package="pendulum", module="pendulum", names=["DateTime", "Timezone"]
        ),
    ]
    file_path: str | Path

    dag_id: DagId
    schedule: str | OrbiterTimetable | None = None
    catchup: bool = False
    start_date: DateTime | datetime = DateTime(1970, 1, 1)
    tags: List[str] = None
    default_args: Dict[str, Any] = dict()
    params: Dict[str, Any] = dict()
    doc_md: str | None = None

    tasks: Dict[str, OrbiterOperator | OrbiterTaskGroup] = dict()

    nullable_attributes: List[str] = ["catchup", "schedule"]
    render_attributes: List[str] = [
        "dag_id",
        "schedule",
        "start_date",
        "catchup",
        "tags",
        "default_args",
        "params",
        "doc_md",
    ]

    def repr(self):
        return (
            f"OrbiterDAG("
            f"dag_id={self.dag_id}, "
            f"schedule={self.schedule}, "
            f"start_date={self.start_date}, "
            f"catchup={self.catchup})"
        )

    # noinspection t
    def __add__(self, other):
        if other.tasks:
            for task in other.tasks.values():
                self.add_tasks(task)
        if other.orbiter_conns:
            for conn in other.orbiter_conns:
                self.orbiter_conns.add(conn)
        if other.orbiter_vars:
            for var in other.orbiter_vars:
                self.orbiter_vars.add(var)
        if other.orbiter_env_vars:
            for env_var in other.orbiter_env_vars:
                self.orbiter_env_vars.add(env_var)
        if other.orbiter_includes:
            for include in other.orbiter_includes:
                self.orbiter_includes.add(include)
        if other.model_extra:
            for key in other.model_extra.keys():
                self.model_extra[key] = self.model_extra[key] or other.model_extra[key]
        for key in self.render_attributes:
            setattr(self, key, getattr(self, key) or getattr(other, key))
        return self

    def _dag_to_ast(self) -> ast.Expr:
        """
        Returns the `DAG(...)` object.
        OrbiterDAG._to_ast will handle the rest (like imports, the context manager, tasks, and task dependencies)

        ```pycon
        >>> from orbiter.ast_helper import render_ast
        >>> render_ast(OrbiterDAG(dag_id="dag_id", file_path="")._dag_to_ast())
        "DAG(dag_id='dag_id', schedule=None, start_date=DateTime(1970, 1, 1, 0, 0, 0), catchup=False)"

        ```

        ```pycon
        >>> render_ast(OrbiterDAG(
        ...    dag_id="dag_id",
        ...    file_path="",
        ...    default_args={},
        ...    params={},
        ...    schedule="@hourly",
        ...    start_date=datetime(2000, 1, 1),
        ...    description="foo"
        ... )._dag_to_ast())
        "DAG(dag_id='dag_id', schedule='@hourly', start_date=datetime.datetime(2000, 1, 1, 0, 0), catchup=False, description='foo')"

        ```
        :return: `DAG(...)` as an ast.Expr
        """  # noqa: E501

        def prop(k):
            # special case - nullable_attributes can be False / None
            if k in self.nullable_attributes:
                return getattr(self, k)
            attr = getattr(self, k, None) or getattr(self.model_extra, k, None)
            return ast.Name(id=attr.__name__) if isinstance(attr, Callable) else attr

        index_map = {v: i for i, v in enumerate(self.render_attributes)}
        rendered_params = {
            k: prop(k)
            for k in self.render_attributes
            if getattr(self, k) or k in self.nullable_attributes
        }
        extra_params = {k: prop(k) for k in (self.model_extra.keys() or [])}
        return py_object(
            name="DAG",
            **dict(
                sorted(
                    rendered_params.items(),
                    key=lambda k_v: index_map[k_v[0]],
                )
            ),
            **extra_params,
        )

    def add_tasks(
        self,
        tasks: (
            OrbiterOperator
            | OrbiterTaskGroup
            | Iterable[OrbiterOperator | OrbiterTaskGroup]
        ),
    ) -> "OrbiterDAG":
        """
        Add one or more [`OrbiterOperators`][orbiter.objects.task.OrbiterOperator] to the DAG

        ```pycon
        >>> from orbiter.objects.operators.empty import OrbiterEmptyOperator
        >>> OrbiterDAG(file_path="", dag_id="foo").add_tasks(OrbiterEmptyOperator(task_id="bar")).tasks
        {'bar': bar_task = EmptyOperator(task_id='bar')}

        >>> OrbiterDAG(file_path="", dag_id="foo").add_tasks([OrbiterEmptyOperator(task_id="bar")]).tasks
        {'bar': bar_task = EmptyOperator(task_id='bar')}

        ```

        !!! tip

            Validation requires a `OrbiterTaskGroup`, `OrbiterOperator` (or subclass), or list of either to be passed
            ```pycon
            >>> # noinspection PyTypeChecker
            ... OrbiterDAG(file_path="", dag_id="foo").add_tasks("bar")
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...
            >>> # noinspection PyTypeChecker
            ... OrbiterDAG(file_path="", dag_id="foo").add_tasks(["bar"])
            ... # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            pydantic_core._pydantic_core.ValidationError: ...

            ```
        :param tasks: List of [OrbiterOperator][orbiter.objects.task.OrbiterOperator], or OrbiterTaskGroup or subclass
        :type tasks: OrbiterOperator | OrbiterTaskGroup | Iterable[OrbiterOperator | OrbiterTaskGroup]
        :return: self
        :rtype: OrbiterProject
        """
        if (
            isinstance(tasks, OrbiterOperator)
            or isinstance(tasks, OrbiterTaskGroup)
            or issubclass(type(tasks), OrbiterOperator)
        ):
            tasks = [tasks]

        for task in tasks:
            try:
                task_id = getattr(task, "task_id", None) or getattr(
                    task, "task_group_id"
                )
            except AttributeError:
                raise AttributeError(
                    f"Task {task} does not have a task_id or task_group_id attribute"
                )
            self.tasks[task_id] = task
        return self

    # noinspection PyProtectedMember
    def _to_ast(self) -> List[ast.stmt]:
        def dedupe_callable(ast_collection):
            items = []
            for item in ast_collection:
                if isinstance(item, list):
                    items.extend(item)
                else:
                    items.append(item)

            seen = set()
            for item in items:
                if isinstance(item, ast.FunctionDef):
                    if item.name in seen:
                        continue
                    seen.add(item.name)
                yield item

        # DAG Imports, e.g. `from airflow import DAG`
        # Task/TaskGroup Imports, e.g. `from airflow.operators.bash import BashOperator`
        pre_imports = list(
            set(self.imports)
            | set(_get_imports_recursively(self.tasks.values()))
            | (
                set(self.schedule.imports)
                if isinstance(self.schedule, OrbiterTimetable)
                else set()
            )
            | reduce(
                # Look for e.g. on_failure_callback in model_extra, get imports, merge them all
                lambda old, item: old
                | set(getattr(self.model_extra.get(item, {}), "imports", set())),
                CALLBACK_KEYS,
                set(),
            )
        )

        imports = [i._to_ast() for i in sorted(pre_imports)]

        # foo = BashOperator(...)
        task_definitions = list(
            dedupe_callable([task._to_ast() for task in self.tasks.values()])
        )

        # foo >> bar
        task_dependencies = [
            task._downstream_to_ast()
            for task in sorted(self.tasks.values())
            if task._downstream_to_ast()
        ]

        # with DAG(...) as dag:
        with_dag = py_with(
            self._dag_to_ast().value, body=task_definitions + task_dependencies
        )
        return [
            *imports,
            with_dag,
        ]


# https://github.com/pydantic/pydantic/issues/8790
OrbiterDAG.add_tasks = validate_call()(OrbiterDAG.add_tasks)


@validate_call
def to_dag_id(dag_id: str) -> str:
    return clean_value(dag_id)



from __future__ import annotations

import ast
from typing import List

from pydantic import BaseModel

from orbiter.ast_helper import py_import, OrbiterASTBase


class OrbiterRequirement(OrbiterASTBase, BaseModel, extra="forbid"):
    """A requirement for a project (e.g. `apache-airflow-providers-google`), and it's representation in the DAG file.

    Renders via the DAG File (as an import statement),
    [`requirements.txt`, and `packages.txt`](https://www.astronomer.io/docs/astro/cli/develop-project#add-python-os-level-packages-and-airflow-providers)

    !!! tip

        If a given requirement has multiple packages required,
        it can be defined as multiple `OrbiterRequirement` objects.

        Example:
        ```python
        OrbiterTask(
            ...,
            imports=[
                OrbiterRequirement(package="apache-airflow-providers-google", ...),
                OrbiterRequirement(package="bigquery", sys_package="mysql", ...),
            ],
        )
        ```

    :param package: e.g. `"apache-airflow-providers-google"`
    :type package: str, optional
    :param module: e.g. `"airflow.providers.google.cloud.operators.bigquery"`, defaults to `None`
    :type module: str, optional
    :param names: e.g. `["BigQueryCreateEmptyDatasetOperator"]`, defaults to `[]`
    :type names: List[str], optional
    :param sys_package: e.g. `"mysql"` - represents a **Debian** system package
    :type sys_package: Set[str], optional
    """  # noqa: E501

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    package: str | None
    module: str | None
    names: List[str] | None
    sys_package: str | None
    --8<-- [end:mermaid-props]
    """

    package: str | None = None
    module: str | None = None
    names: List[str] | None = []
    sys_package: str | None = None

    def __repr__(self):
        return (
            f"OrbiterRequirements("
            f"names=[{','.join(sorted(self.names))}], "
            f"package={self.package}, "
            f"module={self.module}, "
            f"sys_package={self.sys_package})"
        )

    def __hash__(self):
        return hash(
            f"{self.package}{self.module}{''.join(self.names)}{self.sys_package}"
        )

    def _to_ast(self) -> ast.stmt | ast.Module:
        """Render to `from ... import`
        ```pycon
        >>> from orbiter.ast_helper import render_ast
        >>> render_ast(OrbiterRequirement(
        ...     package="apache-airflow", module="airflow.operators.bash", names=["BashOperator"]
        ... )._to_ast())
        'from airflow.operators.bash import BashOperator'

        ```
        """
        return py_import(module=self.module, names=sorted(self.names))



from orbiter.objects import ImportList, OrbiterRequirement
from orbiter.objects.task import OrbiterOperator, RenderAttributes

__mermaid__ = """
--8<-- [start:mermaid-relationships]
OrbiterOperator "implements" <|-- OrbiterSQLExecuteQueryOperator
--8<-- [end:mermaid-relationships]
"""


class OrbiterSQLExecuteQueryOperator(OrbiterOperator):
    """
    An Airflow
    [Generic SQL Operator](https://registry.astronomer.io/providers/apache-airflow-providers-common-sql/versions/1.10.1/modules/SQLExecuteQueryOperator).
    Used to run SQL against any Database.

    ```pycon
    >>> OrbiterSQLExecuteQueryOperator(
    ...   task_id="foo", conn_id='sql', sql="select 1;"
    ... )
    foo_task = SQLExecuteQueryOperator(task_id='foo', conn_id='sql', sql='select 1;')

    ```
    :param task_id: The `task_id` for the operator
    :type task_id: str
    :param conn_id: The SQL connection to utilize.  (Note: use the `**conn_id(...)` utility function)
    :type conn_id: str
    :param sql: The SQL to execute
    :type sql: str
    :param **kwargs: Extra arguments to pass to the operator
    :param **OrbiterBase: [OrbiterBase][orbiter.objects.OrbiterBase] inherited properties
    """  # noqa: E501

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    operator = "SQLExecuteQueryOperator"
    task_id: str
    conn_id: str
    sql: str
    --8<-- [end:mermaid-props]
    """

    imports: ImportList = [
        OrbiterRequirement(
            package="apache-airflow-providers-common-sql",
            module="airflow.providers.common.sql.operators.sql",
            names=["SQLExecuteQueryOperator"],
        )
    ]
    operator: str = "SQLExecuteQueryOperator"
    conn_id: str
    sql: str

    # noinspection Pydantic
    render_attributes: RenderAttributes = OrbiterOperator.render_attributes + [
        "conn_id",
        "sql",
    ]



from typing import ClassVar

from pydantic import BaseModel

from orbiter.objects.airflow_settings import AirflowSettingsRender


class OrbiterVariable(BaseModel, AirflowSettingsRender, extra="forbid"):
    """An Airflow
    [Variable](https://airflow.apache.org/docs/apache-airflow/stable/core-concepts/variables.html),
    rendered to an
    [`airflow_settings.yaml`](https://www.astronomer.io/docs/astro/cli/develop-project#configure-airflow_settingsyaml-local-development-only)
    file.

    ```pycon
    >>> OrbiterVariable(key="foo", value="bar").render()
    {'variable_value': 'bar', 'variable_name': 'foo'}

    ```
    :param key: The key of the variable
    :type key: str
    :param value: The value of the variable
    :type value: str
    """  # noqa: E501

    __mermaid__ = """
    --8<-- [start:mermaid-props]
    key: str
    value: str
    --8<-- [end:mermaid-props]
    """

    render_prefix: ClassVar[str] = "variable_"

    key: str
    value: str

    def __hash__(self):
        return hash(f"{self.model_dump_json()}")

    def render(self) -> dict:
        # airflow_settings.yaml takes 'name' not 'key'
        res = super().render()
        res[self.render_prefix + "name"] = res[self.render_prefix + "key"]
        del res[self.render_prefix + "key"]
        return res



import os
import sys

TRANSLATION_VERSION = os.getenv("ORBITER_TRANSLATION_VERSION", "latest")
"""The version of the translation ruleset to download. This can be overridden."""

ORBITER_TASK_SUFFIX = os.getenv("ORBITER_TASK_SUFFIX", "_task")
"""By default, we add `_task` as a suffix to a task name to prevent name collision issues. This can be overridden."""

LOG_LEVEL = os.environ.get("LOG_LEVEL", "INFO")
"""You can set the log level to DEBUG to see more detailed logs."""

TRIM_LOG_OBJECT_LENGTH = os.getenv("TRIM_LOG_OBJECT_LENGTH", 1000)
"""Trim the (str) length of logged objects to avoid long logs, set to -1 to disable trimming and log full objects."""

KG_ACCOUNT_ID = "3b189b4c-c047-4fdb-9b46-408aa2978330"
RUNNING_AS_BINARY = getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")



from __future__ import annotations

from typing import List, Set

from orbiter.objects import OrbiterRequirement, ImportList
from orbiter.objects.task import RenderAttributes
from orbiter.objects.timetables import OrbiterTimetable
from orbiter.objects.include import OrbiterInclude

TIMETABLE_CONTENTS = '''from typing import Any, Dict, List, Optional

import pendulum
from airflow.exceptions import AirflowTimetableInvalid
from airflow.plugins_manager import AirflowPlugin
from airflow.timetables.base import DagRunInfo, DataInterval, TimeRestriction, Timetable
from croniter import croniter
from pendulum import DateTime, Duration
from pendulum import instance as pendulum_instance
from pendulum import timezone


class MultiCronTimetable(Timetable):
    valid_units = ["minutes", "hours", "days"]

    def __init__(
        self,
        cron_defs: List[str],
        timezone: str = "UTC",
        period_length: int = 0,
        period_unit: str = "hours",
    ):

        self.cron_defs = cron_defs
        self.timezone = timezone
        self.period_length = period_length
        self.period_unit = period_unit

    def __repr__(self) -> str:
        return "MultiCronTimetable({})".format(self.cron_defs)

    def infer_manual_data_interval(self, run_after: DateTime) -> DataInterval:
        """
        Determines date interval for manually triggered runs.
        This is simply (now - period) to now.
        """
        end = run_after
        start = end if (self.period_length == 0) else self.data_period_start(end)

        return DataInterval(start=start, end=end)

    def next_dagrun_info(
        self,
        *,
        last_automated_data_interval: Optional[DataInterval],
        restriction: TimeRestriction,
    ) -> Optional[DagRunInfo]:
        """Determines when the DAG should be scheduled."""

        if restriction.earliest is None:
            # No start_date. Don't schedule.
            return None

        is_first_run = last_automated_data_interval is None

        if is_first_run:
            if restriction.catchup:
                scheduled_time = self.next_scheduled_run_time(restriction.earliest)

            else:
                scheduled_time = self.previous_scheduled_run_time()
                if scheduled_time is None:
                    # No previous cron time matched. Find one in the future.
                    scheduled_time = self.next_scheduled_run_time()
        else:
            last_scheduled_time = last_automated_data_interval.end

            if restriction.catchup:
                scheduled_time = self.next_scheduled_run_time(last_scheduled_time)

            else:
                scheduled_time = self.previous_scheduled_run_time()

                if scheduled_time is None or scheduled_time == last_scheduled_time:
                    # No previous cron time matched,
                    # or the matched cron time was the last execution time,
                    scheduled_time = self.next_scheduled_run_time()

                elif scheduled_time > last_scheduled_time:
                    # Matched cron time was after last execution time, but before now.
                    # Use this cron time
                    pass

                else:
                    # The last execution time is after the most recent matching cron time.
                    # Next scheduled run will be in the future
                    scheduled_time = self.next_scheduled_run_time()

        if scheduled_time is None:
            return None

        if restriction.latest is not None and scheduled_time > restriction.latest:
            # Over the DAG's scheduled end; don't schedule.
            return None

        start = self.data_period_start(scheduled_time)
        return DagRunInfo(
            run_after=scheduled_time,
            data_interval=DataInterval(start=start, end=scheduled_time),
        )

    def data_period_start(self, period_end: DateTime):
        return period_end - Duration(**{self.period_unit: self.period_length})

    def croniter_values(self, base_datetime=None):
        if not base_datetime:
            tz = timezone(self.timezone)
            base_datetime = pendulum.now(tz)

        return [croniter(expr, base_datetime) for expr in self.cron_defs]

    def next_scheduled_run_time(self, base_datetime: DateTime = None):
        min_date = None
        tz = timezone(self.timezone)

        base_datetime_localized = (
            base_datetime.in_timezone(tz) if base_datetime else pendulum.now(tz)
        )

        for cron in self.croniter_values(base_datetime_localized):
            next_date = cron.get_next(DateTime)

            min_date = next_date if not min_date else min(min_date, next_date)

        if min_date is None:
            return None
        return pendulum_instance(min_date)

    def previous_scheduled_run_time(self, base_datetime: DateTime = None):
        """
        Get the most recent time in the past that matches one of the cron schedules
        """
        max_date = None
        tz = timezone(self.timezone)

        base_datetime_localized = (
            base_datetime.in_timezone(tz) if base_datetime else pendulum.now(tz)
        )

        for cron in self.croniter_values(base_datetime_localized):
            prev_date = cron.get_prev(DateTime)

            max_date = prev_date if not max_date else max(max_date, prev_date)

        if max_date is None:
            return None
        return pendulum_instance(max_date)

    def validate(self) -> None:
        if not self.cron_defs:
            raise AirflowTimetableInvalid(
                "At least one cron definition must be present"
            )

        if self.period_unit not in self.valid_units:
            raise AirflowTimetableInvalid(
                f"period_unit must be one of {self.valid_units}"
            )

        if self.period_length < 0:
            raise AirflowTimetableInvalid("period_length must not be less than zero")

        try:
            self.croniter_values()
        except Exception as e:
            raise AirflowTimetableInvalid(str(e)) from None

    @property
    def summary(self) -> str:
        """A short summary for the timetable.

        This is used to display the timetable in the web UI. A cron expression
        timetable, for example, can use this to display the expression.
        """
        return " || ".join(self.cron_defs) + f" [TZ: {self.timezone}]"

    def serialize(self) -> Dict[str, Any]:
        """Serialize the timetable for JSON encoding.

        This is called during DAG serialization to store timetable information
        in the database. This should return a JSON-serializable dict that will
        be fed into ``deserialize`` when the DAG is deserialized.
        """
        return dict(
            cron_defs=self.cron_defs,
            timezone=self.timezone,
            period_length=self.period_length,
            period_unit=self.period_unit,
        )

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> "MultiCronTimetable":
        """Deserialize a timetable from data.

        This is called when a serialized DAG is deserialized. ``data`` will be
        whatever was returned by ``serialize`` during DAG serialization.
        """
        return cls(**data)


class CustomTimetablePlugin(AirflowPlugin):
    name = "multi_cron_timetable"
    timetables = [MultiCronTimetable]
'''


class OrbiterMultiCronTimetable(OrbiterTimetable):
    """
    An
    [Airflow Timetable](https://airflow.apache.org/docs/apache-airflow/stable/authoring-and-scheduling/timetable.html)
    that can be supplied with multiple cron strings.

    ```pycon
    >>> OrbiterMultiCronTimetable(cron_defs=["*/5 * * * *", "*/7 * * * *"])
    MultiCronTimetable(cron_defs=['*/5 * * * *', '*/7 * * * *'])

    ```
    :params cron_defs: A list of cron strings
    :type cron_defs: List[str]
    :params timezone: The timezone to use for the timetable
    :type timezone: str
    :params period_length: The length of the period
    :type period_length: int
    :params period_unit: The unit of the period
    :type period_unit: str
    """

    imports: ImportList = [
        OrbiterRequirement(
            package="croniter",
            module="multi_cron_timetable",
            names=["MultiCronTimetable"],
        )
    ]
    orbiter_includes: Set["OrbiterInclude"] = {
        OrbiterInclude(
            filepath="plugins/multi_cron_timetable.py", contents=TIMETABLE_CONTENTS
        )
    }
    render_attributes: RenderAttributes = [
        "cron_defs",
        "timezone",
        "period_length",
        "period_unit",
    ]

    cron_defs: List[str]
    timezone: str | None = None
    period_length: int | None = None
    period_unit: str | None = None
